from unittest import mock
from unittest.mock import patch

import pandas as pd
import pytest
import synapseclient

import genie_registry.maf
from genie_registry.maf import maf
from genie_registry.mafSP import mafSP

syn = mock.create_autospec(synapseclient.Synapse)

maf_class = maf(syn, "SAGE")
mafsp_class = mafSP(syn, "SAGE")


def test_invalidname_validateFilename():
    with pytest.raises(AssertionError):
        maf_class.validateFilename(['foo'])


def test_valid_validateFilename():
    assert maf_class.validateFilename([
        "data_mutations_extended_SAGE.txt"]) == "maf"


def test_perfect_validation():
    mafDf = pd.DataFrame(dict(
        CHROMOSOME=[1, 2, 3, 4, 5],
        START_POSITION=[1, 2, 3, 4, 2],
        REFERENCE_ALLELE=["A", "A", "A", "A", "A"],
        TUMOR_SAMPLE_BARCODE=["ID1-1", "ID1-1", "ID1-1", "ID1-1", "ID1-1"],
        T_ALT_COUNT=[1, 2, 3, 4, 3],
        T_DEPTH=[1, 2, 3, 4, 3],
        T_REF_COUNT=[1, 2, 3, 4, 3],
        N_DEPTH=[1, 2, 3, float('nan'), 3],
        N_REF_COUNT=[1, 2, 3, 4, 3],
        N_ALT_COUNT=[1, 2, 3, 4, 3],
        TUMOR_SEQ_ALLELE2=["A", "A", "A", "A", "A"]))

    error, warning = maf_class._validate(mafDf)
    assert error == ""
    assert warning == ""
    error, warning = mafsp_class._validate(mafDf)
    assert error == ""
    assert warning == ""


def test_firstcolumn_validation():
    """Tests if first column isn't correct"""
    mafDf = pd.DataFrame({
        'REFERENCE_ALLELE': ["A", "B", "C", "D", "E"],
        "START_POSITION": [1, 2, 3, 4, 2],
        "CHROMOSOME": ["A", "A", "A", "A", "A"],
        "TUMOR_SAMPLE_BARCODE": ["ID1-1", "ID1-1", "ID1-1", "ID1-1", "ID1-1"],
        "T_ALT_COUNT": [1, 2, 3, 4, 3],
        "T_DEPTH": [1, 2, 3, 4, 3],
        "T_REF_COUNT": [1, 2, 3, 4, 3],
        "N_DEPTH": [1, 2, 3, 4, 3],
        "N_REF_COUNT": [1, 2, 3, 4, 3],
        "N_ALT_COUNT": [1, 2, 3, 4, 3],
        "TUMOR_SEQ_ALLELE2": ["A", "A", "A", "A", "A"]})
    order = ['REFERENCE_ALLELE', 'START_POSITION', 'CHROMOSOME',
             'TUMOR_SAMPLE_BARCODE', 'T_ALT_COUNT', 'T_DEPTH',
             'T_REF_COUNT', 'N_DEPTH', 'N_REF_COUNT', 'N_ALT_COUNT',
             'TUMOR_SEQ_ALLELE2']
    error, warning = maf_class._validate(mafDf[order])
    expectedErrors = ("maf: First column header must be "
                      "one of these: CHROMOSOME, HUGO_SYMBOL, "
                      "TUMOR_SAMPLE_BARCODE.\n")
    assert error == expectedErrors
    assert warning == ""


def test_missingcols_validation():
    """Tests missing columns"""
    emptydf = pd.DataFrame()
    error, warning = maf_class._validate(emptydf)
    expected_errors = ("maf: Must at least have these headers: "
                       "CHROMOSOME,START_POSITION,REFERENCE_ALLELE,"
                       "TUMOR_SAMPLE_BARCODE,T_ALT_COUNT,"
                       "TUMOR_SEQ_ALLELE2. "
                       "If you are writing your maf file with R, please make"
                       "sure to specify the 'quote=FALSE' parameter.\n"
                       "maf: If missing T_DEPTH, must have T_REF_COUNT!\n")
    expected_warnings = ("maf: Does not have the column headers "
                         "that can give extra information to the processed "
                         "maf: T_REF_COUNT, N_DEPTH, N_REF_COUNT, "
                         "N_ALT_COUNT.\n")
    assert error == expected_errors
    assert warning == expected_warnings


def test_errors_validation():
    mafDf = pd.DataFrame(dict(
        CHROMOSOME=[1, "chr2", "WT", 4, 5],
        START_POSITION=[1, 2, 3, 4, 2],
        REFERENCE_ALLELE=["NA", float('nan'), "A", "A", "A"],
        TUMOR_SAMPLE_BARCODE=["ID1-1", "ID1-1", "ID1-1", "ID1-1", "ID1-1"],
        TUMOR_SEQ_ALLELE2=["B", "B", "B", "B", "B"],
        T_ALT_COUNT=[1, 2, 3, 4, 3],
        T_DEPTH=[1, 2, 3, 4, 3],
        N_REF_COUNT=[1, 2, 3, 4, 3],
        N_ALT_COUNT=[1, 2, 3, 4, 3]))

    error, warning = maf_class._validate(mafDf)

    expectedErrors = (
        "maf: "
        "REFERENCE_ALLELE can't have any blank or null values.\n"
        "maf: "
        "CHROMOSOME column cannot have any values that start "
        "with 'chr' or any 'WT' values.\n"
    )
    expectedWarnings = ("maf: "
                        "Does not have the column headers that can give "
                        "extra information to the processed maf: "
                        "T_REF_COUNT, N_DEPTH.\n"
                        "maf: "
                        "REFERENCE_ALLELE column contains 'NA' values, "
                        "which cannot be placeholders for blank values.  "
                        "Please put in empty strings for blank values.\n")

    assert error == expectedErrors
    assert warning == expectedWarnings


def test_invalid_validation():
    mafDf = pd.DataFrame(dict(
        CHROMOSOME=[1, 2, 3, 4, 2, 4],
        T_ALT_COUNT=[1, 2, 3, 4, 3, 4],
        START_POSITION=[1, 2, 3, 4, 2, 4],
        REFERENCE_ALLELE=["A ", "A ", "A", "A ", "A ", " A"],
        TUMOR_SAMPLE_BARCODE=["ID1-1", "ID1-1", "ID1-1",
                              "ID1-1", "ID1-1", "ID1-1"],
        N_DEPTH=[1, 2, 3, 4, 3, 4],
        N_REF_COUNT=[1, 2, 3, "string", 3, 4],
        N_ALT_COUNT=[1, 2, 3, 4, 3, 4],
        TUMOR_SEQ_ALLELE2=["NA", float('nan'), " A", "A ", " A", " A"]))

    with patch.object(genie_registry.maf, "_check_tsa1_tsa2",
                      return_value="") as check_tsa1_tsa2:
        error, warning = maf_class._validate(mafDf)
        check_tsa1_tsa2.assert_called_once_with(mafDf)
    expectedErrors = (
        "maf: Must not have duplicated variants. "
        "Samples with duplicated variants: ID1-1\n"
        "maf: "
        "If missing T_DEPTH, must have T_REF_COUNT!\n"
        "maf: N_REF_COUNT must be a numerical column.\n"
        "maf: "
        "TUMOR_SEQ_ALLELE2 can't have any blank or null values.\n"
    )
    expectedWarnings = (
        "maf: TUMOR_SEQ_ALLELE2 column contains 'NA' values, "
        "which cannot be placeholders for blank values.  "
        "Please put in empty strings for blank values.\n"
        "maf: Does not have the column headers that can give "
        "extra information to the processed maf: T_REF_COUNT.\n")
    assert error == expectedErrors
    assert warning == expectedWarnings


@pytest.mark.parametrize("col", ["temp", "REFERENCE_ALLELE"])
def test_noerror__check_allele_col(col):
    """Test error and warning is an empty string if REF col isn't passed in"""
    df = pd.DataFrame(dict(
        REFERENCE_ALLELE=["A", "A"]
    ))
    error, warning = genie_registry.maf._check_allele_col(df, col)
    assert error == ""
    assert warning == ""


def test_warning__check_allele_col():
    """Test warning occurs when 'NA' string is passed in"""
    df = pd.DataFrame(dict(
        TEMP=["NA", "A"]
    ))
    error, warning = genie_registry.maf._check_allele_col(df, "TEMP")
    assert error == ""
    assert warning == (
        "maf: "
        "TEMP column contains 'NA' values, "
        "which cannot be placeholders for blank values.  "
        "Please put in empty strings for blank values.\n"
    )


def test_error__check_allele_col():
    """Test error occurs when blank allele is passed in"""
    df = pd.DataFrame(dict(
        TEMP=[float('nan'), "A"]
    ))
    error, warning = genie_registry.maf._check_allele_col(df, "TEMP")
    assert error == (
        "maf: TEMP can't have any blank or null values.\n"
    )
    assert warning == ""


def test_invalid__check_tsa1_tsa2():
    """Test the scenario in which maf file has TSA1 and TSA2 and fails"""
    df = pd.DataFrame(dict(
        REFERENCE_ALLELE=["A", "A", "A"],
        TUMOR_SEQ_ALLELE1=["B", "B", "B"],
        TUMOR_SEQ_ALLELE2=["C", "C", "C"]
    ))
    error = genie_registry.maf._check_tsa1_tsa2(df)
    assert error == (
        "maf: Contains both "
        "TUMOR_SEQ_ALLELE1 and TUMOR_SEQ_ALLELE2 columns. "
        "The values in TUMOR_SEQ_ALLELE1 must be the same as "
        "all the values in REFERENCE_ALELLE OR TUMOR_SEQ_ALLELE2."
    )


@pytest.mark.parametrize(
    "df",
    [pd.DataFrame(dict(REFERENCE_ALLELE=["A", "A", "A"],
                       TUMOR_SEQ_ALLELE1=["C", "C", "C"],
                       TUMOR_SEQ_ALLELE2=["C", "C", "C"])),
     pd.DataFrame(dict(REFERENCE_ALLELE=["C", "C", "C"],
                       TUMOR_SEQ_ALLELE1=["C", "C", "C"],
                       TUMOR_SEQ_ALLELE2=["A", "A", "A"]))]
)
def test_valid__check_tsa1_tsa2(df):
    """Test valid TSA1 and TSA2"""
    error = genie_registry.maf._check_tsa1_tsa2(df)
    assert error == ""



import datetime
import json
from unittest import mock

import pandas as pd
import pytest
import synapseclient

import genie_registry
from genie_registry.clinical import clinical


def createMockTable(dataframe):
    table = mock.create_autospec(synapseclient.table.CsvFileTable)
    table.asDataFrame.return_value = dataframe
    return(table)


def table_query_results(*args):
    return(table_query_results_map[args])


no_nan = pd.DataFrame(dict(
    CODE=[1, 2, 3, 4, 99],
    CBIO_LABEL=['Test', 'Why', 'foo', 'Me', 'Unknown'],
    DESCRIPTION=['non', 'asdf', 'asdf', 'asdff', 'asdfasdf']))

sexdf = pd.DataFrame(dict(
    CODE=[1, 2, 99],
    CBIO_LABEL=['Male', 'Female', 'Unknown'],
    DESCRIPTION=['Male', 'Female', 'Not coded']))

table_query_results_map = {
    ("SELECT * FROM syn7434222",): createMockTable(sexdf),
    ("SELECT * FROM syn7434236",): createMockTable(no_nan),
    ("SELECT * FROM syn7434242",): createMockTable(no_nan),
    ("SELECT * FROM syn7434273",): createMockTable(no_nan)}

syn = mock.create_autospec(synapseclient.Synapse)
syn.tableQuery.side_effect = table_query_results
clin_class = clinical(syn, "SAGE")

# oncotree_url = \
#  'http://oncotree.mskcc.org/api/tumor_types.txt?version=oncotree_latest_stable'
json_oncotreeurl = \
 "http://oncotree.mskcc.org/api/tumorTypes/tree?version=oncotree_2017_06_21"

onco_map_dict = {
  "AMPCA": {
    'CANCER_TYPE': "Ampullary Cancer",
    'CANCER_TYPE_DETAILED': "Ampullary Carcinoma",
    'ONCOTREE_PRIMARY_NODE': "AMPULLA_OF_VATER",
    'ONCOTREE_SECONDARY_NODE': "AMPCA"},
  "TESTIS": {
    'CANCER_TYPE': "Testicular Cancer, NOS",
    'CANCER_TYPE_DETAILED': "Testis",
    'ONCOTREE_PRIMARY_NODE': "TESTIS",
    'ONCOTREE_SECONDARY_NODE': ''},
  "UCEC": {
    'CANCER_TYPE': "Endometrial Cancer",
    'CANCER_TYPE_DETAILED': "Endometrial Carcinoma",
    'ONCOTREE_PRIMARY_NODE': "UTERUS",
    'ONCOTREE_SECONDARY_NODE': "UCEC"}}


def test_filetype():
    assert clin_class._fileType == "clinical"


@pytest.fixture(params=[
    (["foo"]),
    (["foo", "data_clinical_supp_sample_SAGE.txt"])
    ])
def filename_fileformat_map(request):
    return request.param


def test_incorrect_validatefilename(filename_fileformat_map):
    filepath_list = filename_fileformat_map
    with pytest.raises(AssertionError):
        clin_class.validateFilename(filepath_list)


def test_correct_validatefilename():
    assert clin_class.validateFilename(
        ["data_clinical_supp_SAGE.txt"]) == "clinical"
    assert clin_class.validateFilename(
        ["data_clinical_supp_sample_SAGE.txt",
         "data_clinical_supp_patient_SAGE.txt"]) == "clinical"


def test_patient_fillvs__process():
    '''
    Test filling out of vital status values
    This will be removed once vital status values are required
    - capitalized column headers
    - remapping of values
    - Fill out CENTER column
    - Append GENIE-CENTER-..
    '''
    expected_patientdf = pd.DataFrame(dict(
        PATIENT_ID=["GENIE-SAGE-ID1", "GENIE-SAGE-ID2", "GENIE-SAGE-ID3",
                    "GENIE-SAGE-ID4", "GENIE-SAGE-ID5"],
        SEX=['Male', 'Female', 'Male', 'Female', 'Unknown'],
        PRIMARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        SECONDARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        TERTIARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        ETHNICITY=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        BIRTH_YEAR=[1990, 1990, 1990, 1990, 1990],
        CENTER=["SAGE", "SAGE", "SAGE", "SAGE", "SAGE"],
        INT_DOD=['Not Collected', 'Not Collected', 'Not Collected',
                 'Not Collected', 'Not Collected'],
        INT_CONTACT=['Not Collected', 'Not Collected', 'Not Collected',
                     'Not Collected', 'Not Collected'],
        DEAD=['Not Collected', 'Not Collected', 'Not Collected',
              'Not Collected', 'Not Collected'],
        YEAR_DEATH=['Not Collected', 'Not Collected', 'Not Collected',
                    'Not Collected', 'Not Collected'],
        YEAR_CONTACT=['Not Collected', 'Not Collected', 'Not Collected',
                      'Not Collected', 'Not Collected']))

    patientdf = pd.DataFrame(dict(
        PATIENT_Id=["ID1", "ID2", "ID3", "ID4", "ID5"],
        sex=[1, 2, 1, 2, 99],
        PRIMARY_RACE=[1, 2, 3, 4, 99],
        Secondary_RACE=[1, 2, 3, 4, 99],
        TERTIARY_RACE=[1, 2, 3, 4, 99],
        ETHNICITY=[1, 2, 3, 4, 99],
        BIRTH_YEAR=[1990, 1990, 1990, 1990, 1990],
        CENTER=["FOO", "FOO", "FOO", "FOO", "FOO"]))
    patient_cols = [
        "PATIENT_ID", "SEX", "PRIMARY_RACE", "SECONDARY_RACE",
        "TERTIARY_RACE", "ETHNICITY", "BIRTH_YEAR", "CENTER",
        'YEAR_CONTACT', 'YEAR_DEATH', 'INT_CONTACT', 'INT_DOD', 'DEAD']
    clinical_template = pd.DataFrame(columns=patient_cols)
    new_patientdf = clin_class._process(patientdf, clinical_template)
    assert new_patientdf.columns.isin(expected_patientdf.columns).all()
    assert expected_patientdf.equals(new_patientdf[expected_patientdf.columns])


def test_patient_lesscoltemplate__process():
    '''
    Test scope is excluding values.
    Only those value defined by the scope will be written out
    '''
    expected_patientdf = pd.DataFrame(dict(
        PATIENT_ID=["GENIE-SAGE-ID1", "GENIE-SAGE-ID2", "GENIE-SAGE-ID3",
                    "GENIE-SAGE-ID4", "GENIE-SAGE-ID5"],
        SEX=['Male', 'Female', 'Male', 'Female', 'Unknown'],
        PRIMARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        SECONDARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        TERTIARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        ETHNICITY=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        BIRTH_YEAR=[1990, 1990, 1990, 1990, 1990],
        CENTER=["SAGE", "SAGE", "SAGE", "SAGE", "SAGE"],
        INT_DOD=['Not Collected', 'Not Collected', 'Not Collected',
                 'Not Collected', 'Not Collected'],
        INT_CONTACT=['Not Collected', 'Not Collected', 'Not Collected',
                     'Not Collected', 'Not Collected'],
        DEAD=['Not Collected', 'Not Collected', 'Not Collected',
              'Not Collected', 'Not Collected'],
        YEAR_DEATH=['Not Collected', 'Not Collected', 'Not Collected',
                    'Not Collected', 'Not Collected'],
        YEAR_CONTACT=['Not Collected', 'Not Collected', 'Not Collected',
                      'Not Collected', 'Not Collected']))
    # TEST patient processing
    patientdf = pd.DataFrame(dict(
        PATIENT_Id=["ID1", "ID2", "ID3", "ID4", "ID5"],
        sex=[1, 2, 1, 2, 99],
        PRIMARY_RACE=[1, 2, 3, 4, 99],
        Secondary_RACE=[1, 2, 3, 4, 99],
        TERTIARY_RACE=[1, 2, 3, 4, 99],
        ETHNICITY=[1, 2, 3, 4, 99],
        BIRTH_YEAR=[1990, 1990, 1990, 1990, 1990],
        CENTER=["FOO", "FOO", "FOO", "FOO", "FOO"]))
    patient_cols = [
        "PATIENT_ID", "SEX", "PRIMARY_RACE", "SECONDARY_RACE",
        "TERTIARY_RACE", "ETHNICITY", "BIRTH_YEAR", "CENTER",
        'YEAR_CONTACT', 'YEAR_DEATH']
    clinical_template = pd.DataFrame(columns=patient_cols)
    new_patientdf = clin_class._process(patientdf, clinical_template)

    assert new_patientdf.columns.isin(patient_cols).all()
    assert expected_patientdf[patient_cols].equals(new_patientdf[patient_cols])


def test_patient_fillcols__process():
    '''
    Filling in of RACE/ETHNITICY columns as some centers don't require them
    '''
    expected_patientdf = pd.DataFrame(dict(
        PATIENT_ID=["GENIE-SAGE-ID1", "GENIE-SAGE-ID2", "GENIE-SAGE-ID3",
                    "GENIE-SAGE-ID4", "GENIE-SAGE-ID5"],
        SEX=['Male', 'Female', 'Male', 'Female', 'Unknown'],
        PRIMARY_RACE=['Not Collected', 'Not Collected', 'Not Collected',
                      'Not Collected', 'Not Collected'],
        SECONDARY_RACE=['Not Collected', 'Not Collected', 'Not Collected',
                        'Not Collected', 'Not Collected'],
        TERTIARY_RACE=['Not Collected', 'Not Collected', 'Not Collected',
                       'Not Collected', 'Not Collected'],
        ETHNICITY=['Not Collected', 'Not Collected', 'Not Collected',
                   'Not Collected', 'Not Collected'],
        BIRTH_YEAR=[1990, 1990, 1990, 1990, 1990],
        CENTER=["SAGE", "SAGE", "SAGE", "SAGE", "SAGE"],
        INT_DOD=['Not Collected', 'Not Collected', 'Not Collected',
                 'Not Collected', 'Not Collected'],
        INT_CONTACT=['Not Collected', 'Not Collected', 'Not Collected',
                     'Not Collected', 'Not Collected'],
        DEAD=['Not Collected', 'Not Collected', 'Not Collected',
              'Not Collected', 'Not Collected'],
        YEAR_DEATH=['Not Collected', 'Not Collected', 'Not Collected',
                    'Not Collected', 'Not Collected'],
        YEAR_CONTACT=['Not Collected', 'Not Collected', 'Not Collected',
                      'Not Collected', 'Not Collected']))
    # TEST patient processing
    # Clinical file headers are capitalized prior to processing
    patientdf = pd.DataFrame(dict(
        PATIENT_Id=["ID1", "ID2", "ID3", "ID4", "ID5"],
        sex=[1, 2, 1, 2, 99],
        BIRTH_YEAR=[1990, 1990, 1990, 1990, 1990],
        CENTER=["FOO", "FOO", "FOO", "FOO", "FOO"]))
    patient_cols = [
        "PATIENT_ID", "SEX", "PRIMARY_RACE", "SECONDARY_RACE",
        "TERTIARY_RACE", "ETHNICITY", "BIRTH_YEAR", "CENTER",
        'YEAR_CONTACT', 'YEAR_DEATH', 'INT_CONTACT', 'INT_DOD', 'DEAD']
    clinical_template = pd.DataFrame(columns=patient_cols)
    new_patientdf = clin_class._process(patientdf, clinical_template)
    assert new_patientdf.columns.isin(expected_patientdf.columns).all()
    assert expected_patientdf.equals(new_patientdf[expected_patientdf.columns])


def test_patient_vs__process():
    '''
    Test vital status columns being propogated with same data
    '''
    expected_patientdf = pd.DataFrame(dict(
        PATIENT_ID=["GENIE-SAGE-ID1", "GENIE-SAGE-ID2", "GENIE-SAGE-ID3",
                    "GENIE-SAGE-ID4", "GENIE-SAGE-ID5"],
        SEX=['Male', 'Female', 'Male', 'Female', 'Unknown'],
        PRIMARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        SECONDARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        TERTIARY_RACE=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        ETHNICITY=['Test', 'Why', 'foo', 'Me', 'Unknown'],
        BIRTH_YEAR=[1990, 1990, 1990, 1990, 1990],
        CENTER=["SAGE", "SAGE", "SAGE", "SAGE", "SAGE"],
        YEAR_DEATH=["Unknown", "Not Collected", "Not Applicable", 1990, 1990],
        YEAR_CONTACT=["Unknown", "Not Collected", 1990, 1990, 1990],
        INT_CONTACT=['>32485', '<6570', 'Unknown', 'Not Collected', 2000],
        INT_DOD=['>32485', '<6570', 'Unknown', 'Not Collected',
                 'Not Applicable'],
        DEAD=[True, False, 'Unknown', 'Not Collected', True]))
    # TEST patient processing
    # Clinical file headers are capitalized prior to processing
    patientdf = pd.DataFrame(dict(
        PATIENT_Id=["ID1", "ID2", "ID3", "ID4", "ID5"],
        sex=[1, 2, 1, 2, 99],
        PRIMARY_RACE=[1, 2, 3, 4, 99],
        Secondary_RACE=[1, 2, 3, 4, 99],
        TERTIARY_RACE=[1, 2, 3, 4, 99],
        ETHNICITY=[1, 2, 3, 4, 99],
        BIRTH_YEAR=[1990, 1990, 1990, 1990, 1990],
        CENTER=["FOO", "FOO", "FOO", "FOO", "FOO"],
        YEAR_DEATH=["Unknown", "Not Collected", "Not Applicable",
                    1990, 1990],
        YEAR_CONTACT=["Unknown", "Not Collected", 1990, 1990, 1990],
        INT_CONTACT=['>32485', '<6570', 'Unknown', 'Not Collected', 2000],
        INT_DOD=['>32485', '<6570', 'Unknown', 'Not Collected',
                 'Not Applicable'],
        DEAD=[True, False, 'Unknown', 'Not Collected', True]))
    patient_cols = [
        "PATIENT_ID", "SEX", "PRIMARY_RACE", "SECONDARY_RACE",
        "TERTIARY_RACE", "ETHNICITY", "BIRTH_YEAR", "CENTER",
        'YEAR_CONTACT', 'YEAR_DEATH', 'INT_CONTACT', 'INT_DOD', 'DEAD']
    clinical_template = pd.DataFrame(columns=patient_cols)
    new_patientdf = clin_class._process(patientdf, clinical_template)
    assert expected_patientdf.equals(new_patientdf[expected_patientdf.columns])


def test_sample__process():
    '''
    Test sample processing
    - column headers are capitalized
    - SEQ_DATE is normalized (Mon-YYYY, Release)
    - Allow UNKNOWN oncotree value
    - Add on GENIE-CENTER-...
    - Remapping of SAMPLE_TYPE/SAMPLE_TYPE_DETAILED value
    - SEQ_YEAR from SEQ_DATE, nan if SEQ_DATE is Release

    '''
    expected_sampledf = pd.DataFrame(dict(
        SAMPLE_ID=["GENIE-SAGE-ID1-1", "GENIE-SAGE-ID2-1", "GENIE-SAGE-ID3-1",
                   "GENIE-SAGE-ID4-1", "GENIE-SAGE-ID5-1"],
        PATIENT_ID=["GENIE-SAGE-ID1", "GENIE-SAGE-ID2", "GENIE-SAGE-ID3",
                    "GENIE-SAGE-ID4", "GENIE-SAGE-ID5"],
        AGE_AT_SEQ_REPORT=[100000, 100000, 100000, 100000, 100000],
        ONCOTREE_CODE=['AMPCA', 'UNKNOWN', 'AMPCA', 'AMPCA', 'AMPCA'],
        SAMPLE_TYPE=['Test', 'Why', 'foo', 'Me', 'Me'],
        CENTER=["SAGE", "SAGE", "SAGE", "SAGE", "SAGE"],
        SAMPLE_TYPE_DETAILED=['non', 'asdf', 'asdf', 'asdff', 'asdff'],
        SEQ_ASSAY_ID=['SAGE-1', 'SAGE-1', 'SAGE-1', 'SAGE-1', 'SAGE-1'],
        SEQ_DATE=['Jan-2012', 'Apr-2013', 'Jul-2014', 'Oct-2015', 'Release'],
        SEQ_YEAR=[2012, 2013, 2014, 2015, float('nan')]))
    sample_cols = [
        "SAMPLE_ID", "PATIENT_ID", "AGE_AT_SEQ_REPORT", "ONCOTREE_CODE",
        "SAMPLE_TYPE", "SEQ_ASSAY_ID", 'SEQ_DATE', 'SAMPLE_TYPE_DETAILED',
        'SEQ_YEAR']

    clinical_template = pd.DataFrame(columns=sample_cols)
    # patient = False
    sampledf = pd.DataFrame(dict(
        SAMPLE_ID=["ID1-1", "ID2-1", "ID3-1", "ID4-1", "ID5-1"],
        PATIENT_ID=["ID1", "ID2", "ID3", "ID4", "ID5"],
        Age_AT_SEQ_REPORT=[100000, 100000, 100000, 100000, 100000],
        ONCOTree_CODE=['AMPCA', ' UNKNOWN', 'AMPCA', 'AMPCA', 'AMPCA'],
        SAMPLE_TYPE=[1, 2, 3, 4, 4],
        SEQ_ASSAY_ID=['SAGE-1', 'SAGE-1', 'SAGE-1', 'SAGE-1', 'SAGE-1'],
        SEQ_DATE=['Jan-2012', 'Apr-2013', 'JUL-2014', 'Oct-2015', 'release']))

    new_sampledf = clin_class._process(sampledf, clinical_template)
    assert new_sampledf.columns.isin(expected_sampledf.columns).all()
    assert expected_sampledf.equals(new_sampledf[expected_sampledf.columns])


def test_perfect__validate():
    '''
    Test perfect validation
    '''
    patientdf = pd.DataFrame(dict(
        PATIENT_ID=["ID1", "ID2", "ID3", "ID4", "ID5"],
        SEX=[1, 2, 1, 2, 99],
        PRIMARY_RACE=[1, 2, 3, 4, 99],
        SECONDARY_RACE=[1, 2, 3, 4, 99],
        TERTIARY_RACE=[1, 2, 3, 4, 99],
        ETHNICITY=[1, 2, 3, 4, 99],
        BIRTH_YEAR=[1222, "Unknown", 1920, 1990, 1990],
        CENTER=["FOO", "FOO", "FOO", "FOO", "FOO"],
        YEAR_DEATH=["Unknown", "Not Collected", "Not Applicable", 1990, 1990],
        YEAR_CONTACT=["Unknown", "Not Collected", 1990, 1990, 1990],
        INT_CONTACT=['>32485', '<6570', 'Unknown', 'Not Collected', 2000],
        INT_DOD=['>32485', '<6570', 'Unknown',
                 'Not Collected', 'Not Applicable'],
        DEAD=[True, False, 'Unknown', 'Not Collected', True]))

    sampledf = pd.DataFrame(dict(
        SAMPLE_ID=["ID1-1", "ID2-1", "ID3-1", "ID4-1", "ID5-1"],
        PATIENT_ID=["ID1", "ID2", "ID3", "ID4", "ID5"],
        AGE_AT_SEQ_REPORT=[">32485", "Unknown", "<6570", 20000, 100000],
        ONCOTREE_CODE=['AMPCA', 'AMPCA', 'Unknown', 'AMPCA', 'AMPCA'],
        SAMPLE_TYPE=[1, 2, 3, 4, 4],
        SEQ_ASSAY_ID=['SAGE-1-1', 'SAGE-SAGE-1', 'SAGE-1', 'SAGE-1', 'SAGE-1'],
        SEQ_DATE=['Jan-2013', 'ApR-2013', 'Jul-2013', 'Oct-2013', 'release']))

    clinicaldf = patientdf.merge(sampledf, on="PATIENT_ID")
    with mock.patch(
            "genie.process_functions.get_oncotree_code_mappings",
            return_value=onco_map_dict) as mock_get_onco_map:
        error, warning = clin_class._validate(clinicaldf, json_oncotreeurl)
        mock_get_onco_map.called_once_with(json_oncotreeurl)
        assert error == ""
        assert warning == ""


def test_nonull__validate():
    '''
    Test that no null values are allowed in the clinical dataframe
    '''
    patientdf = pd.DataFrame(dict(
        PATIENT_ID=["ID1", "ID2", "ID3", "ID4", "ID5"],
        SEX=[1, 2, 1, 2, float('nan')],
        PRIMARY_RACE=[1, 2, 3, 4, float('nan')],
        SECONDARY_RACE=[1, 2, 3, 4, float('nan')],
        TERTIARY_RACE=[1, 2, 3, 4, float('nan')],
        ETHNICITY=[1, 2, 3, 4, float('nan')],
        BIRTH_YEAR=[float('nan'), "Unknown", 1920, 1990, 1990],
        CENTER=["FOO", "FOO", "FOO", "FOO", "FOO"],
        YEAR_DEATH=["Unknown", "Not Collected", float('nan'), 1990, 1990],
        YEAR_CONTACT=["Unknown", "Not Collected", float('nan'), 1990, 1990],
        INT_CONTACT=['>32485', '<6570', 'Unknown', float('nan'), 2000],
        INT_DOD=['>32485', '<6570', 'Unknown',
                 'Not Collected', float('nan')],
        DEAD=[True, False, float('nan'), 'Not Collected', True]))

    sampledf = pd.DataFrame(dict(
        SAMPLE_ID=["ID1-1", "ID2-1", "ID3-1", "ID4-1", "ID5-1"],
        PATIENT_ID=["ID1", "ID2", "ID3", "ID4", "ID5"],
        AGE_AT_SEQ_REPORT=[100000, "Unknown", 20000, float('nan'), 100000],
        ONCOTREE_CODE=['AMPCA', 'AMPCA', 'Unknown', 'AMPCA', 'AMPCA'],
        SAMPLE_TYPE=[1, 2, 3, 4, float('nan')],
        SEQ_ASSAY_ID=['SAGE-1-1', 'SAGE-SAGE-1', 'SAGE-1', 'SAGE-1', 'SAGE-1'],
        SEQ_DATE=['Jan-2013', 'ApR-2013', 'Jul-2013', 'Oct-2013', 'release']))

    clinicaldf = patientdf.merge(sampledf, on="PATIENT_ID")
    with mock.patch(
            "genie.process_functions.get_oncotree_code_mappings",
            return_value=onco_map_dict) as mock_get_onco_map:
        error, warning = clin_class._validate(clinicaldf, json_oncotreeurl)
        mock_get_onco_map.called_once_with(json_oncotreeurl)
        expected_errors = (
            "Sample Clinical File: Please double check your "
            "AGE_AT_SEQ_REPORT. It must be an integer, 'Unknown', "
            "'>32485', '<6570'.\n"
            "Sample Clinical File: Please double check your SAMPLE_TYPE "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
            "Patient Clinical File: Please double check your BIRTH_YEAR "
            "column, it must be an integer in YYYY format <= {year} or "
            "'Unknown', '>89', '<18'.\n"
            "Patient Clinical File: Please double check your YEAR_DEATH "
            "column, it must be an integer in YYYY format <= {year} or "
            "'Unknown', 'Not Collected', 'Not Applicable', "
            "'Not Released', '>89', '<18'.\n"
            "Patient Clinical File: Please double check your YEAR_CONTACT "
            "column, it must be an integer in YYYY format <= {year} or "
            "'Unknown', 'Not Collected', 'Not Released', '>89', '<18'.\n"
            "Patient Clinical File: Please double check your INT_CONTACT "
            "column, it must be an integer, '>32485', '<6570', 'Unknown', "
            "'Not Released' or 'Not Collected'.\n"
            "Patient Clinical File: Please double check your INT_DOD "
            "column, it must be an integer, '>32485', '<6570', 'Unknown', "
            "'Not Collected', 'Not Released' or 'Not Applicable'.\n"
            "Patient Clinical File: Please double check your DEAD column, "
            "it must be True, False, 'Unknown', "
            "'Not Released' or 'Not Collected'.\n"
            "Patient Clinical File: Please double check your PRIMARY_RACE "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
            "Patient Clinical File: Please double check your SECONDARY_RACE "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
            "Patient Clinical File: Please double check your TERTIARY_RACE "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
            "Patient Clinical File: Please double check your SEX column.  "
            "This column must only be these values: 1, 2, 99\n"
            "Patient Clinical File: Please double check your ETHNICITY "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
        )
        assert error == expected_errors.format(
            year=datetime.datetime.utcnow().year
        )
        assert warning == ""


def test_missingcols__validate():
    '''
    Test for missing column errors
    '''
    clinicaldf = pd.DataFrame()
    with mock.patch(
            "genie.process_functions.get_oncotree_code_mappings",
            return_value=onco_map_dict) as mock_get_onco_map:
        error, warning = clin_class._validate(clinicaldf, json_oncotreeurl)
        mock_get_onco_map.called_once_with(json_oncotreeurl)
        expected_errors = (
            "Sample Clinical File: Must have SAMPLE_ID column.\n"
            "Patient Clinical File: Must have PATIENT_ID column.\n"
            "Sample Clinical File: Must have AGE_AT_SEQ_REPORT column.\n"
            "Sample Clinical File: Must have ONCOTREE_CODE column.\n"
            "Sample Clinical File: Must have SAMPLE_TYPE column.\n"
            "Sample Clinical File: Must have SEQ_ASSAY_ID column.\n"
            "Sample Clinical File: Must have SEQ_DATE column.\n"
            "Patient Clinical File: Must have BIRTH_YEAR column.\n"
            "Patient Clinical File: Must have YEAR_DEATH column.\n"
            "Patient Clinical File: Must have YEAR_CONTACT column.\n"
            "Patient Clinical File: Must have INT_CONTACT column.\n"
            "Patient Clinical File: Must have INT_DOD column.\n"
            "Patient Clinical File: Must have DEAD column.\n"
            "Patient Clinical File: Must have SEX column.\n")

        expected_warnings = (
            "Patient Clinical File: Doesn't have PRIMARY_RACE column. "
            "This column will be added\n"
            "Patient Clinical File: Doesn't have SECONDARY_RACE column. "
            "This column will be added\n"
            "Patient Clinical File: Doesn't have TERTIARY_RACE column. "
            "This column will be added\n"
            "Patient Clinical File: Doesn't have ETHNICITY column. "
            "This column will be added\n")
        assert error == expected_errors
        assert warning == expected_warnings


def test_errors__validate():
    '''
    Test for validation errors
    '''
    sampleDf = pd.DataFrame(dict(
        SAMPLE_ID=[float('nan'), "ID2-1", "ID3-1", "ID4-1", "ID5-1"],
        PATIENT_ID=["ID6", "ID2", "ID3", float('nan'), "ID5"],
        AGE_AT_SEQ_REPORT=[10, 100000, ">doo", 100000, 100000],
        ONCOTREE_CODE=['AMPCAD', 'TESTIS', 'AMPCA', 'AMPCA', 'UCEC'],
        SAMPLE_TYPE=[1, 2, 3, 4, 6],
        SEQ_ASSAY_ID=[float('nan'), 'Sage-1', 'SAGE-1', 'S-SAGE-1', 'SAGE-1'],
        SEQ_DATE=['Jane-2013', 'Jan-2013', 'Jan-2013', 'Jan-2013', 'Jan-2013'],
        YEAR_DEATH=["Unknown", "Not Collected", "Not Applicable", 19930, 1990],
        YEAR_CONTACT=["Unknown", "Not Collected", 1990, 1990, 19940],
        INT_CONTACT=['>32485', '<6570', 'Unknown', 'Not Collected', ">foobar"],
        INT_DOD=['>32485', '<6570', 'Unknown', 'Not Collected', '<dense'],
        DEAD=[1, False, 'Unknown', 'Not Collected', 'Not Applicable']))

    patientDf = pd.DataFrame(dict(
        PATIENT_ID=["ID6", "ID2", "ID3", float("nan"), "ID5"],
        SEX=[1, 2, 1, 5, float('nan')],
        PRIMARY_RACE=[1, 2, 3, 6, float('nan')],
        SECONDARY_RACE=[1, 2, 3, 6, float('nan')],
        TERTIARY_RACE=[1, 2, 3, 6, float('nan')],
        ETHNICITY=[1, 2, 3, 6, float('nan')],
        BIRTH_YEAR=[1990, 1990, ">90", 1990, 1990],
        CENTER=["FOO", "FOO", "FOO", "FOO", "FOO"]))
    clinicalDf = patientDf.merge(sampleDf, on="PATIENT_ID")
    with mock.patch(
            "genie.process_functions.get_oncotree_code_mappings",
            return_value=onco_map_dict) as mock_get_onco_map:
        error, warning = clin_class._validate(clinicalDf, json_oncotreeurl)
        mock_get_onco_map.called_once_with(json_oncotreeurl)

        expectedErrors = (
            "Sample Clinical File: PATIENT_ID's much be contained in the "
            "SAMPLE_ID's (ex. SAGE-1 <-> SAGE-1-2)\n"
            "Patient Clinical File: All samples must have associated patient "
            "information and no null patient ids allowed. "
            "These samples are missing patient data: ID4-1\n"
            "Sample Clinical File: Please double check your "
            "AGE_AT_SEQ_REPORT. It must be an integer, 'Unknown', "
            "'>32485', '<6570'.\n"
            "Sample Clinical File: Please double check that all your "
            "ONCOTREE CODES exist in the mapping. You have 1 samples that "
            "don't map. These are the codes that don't map: AMPCAD\n"
            "Sample Clinical File: Please double check your SAMPLE_TYPE "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
            "Sample Clinical File: Please double check your SEQ_ASSAY_ID "
            "columns, there are empty rows.\n"
            "Sample Clinical File: Please make sure your SEQ_ASSAY_IDs start "
            "with your center abbreviation: S-SAGE-1.\n"
            "Sample Clinical File: SEQ_DATE must be one of five values- "
            "For Jan-March: use Jan-YEAR. "
            "For Apr-June: use Apr-YEAR. "
            "For July-Sep: use Jul-YEAR. "
            "For Oct-Dec: use Oct-YEAR. (ie. Apr-2017) "
            "For values that don't have SEQ_DATES that you want "
            "released use 'release'.\n"
            "Patient Clinical File: Please double check your BIRTH_YEAR "
            "column, it must be an integer in YYYY format <= {year} or "
            "'Unknown', '>89', '<18'.\n"
            "Patient Clinical File: Please double check your YEAR_DEATH "
            "column, it must be an integer in YYYY format <= {year} or "
            "'Unknown', 'Not Collected', 'Not Applicable', "
            "'Not Released', '>89', '<18'.\n"
            "Patient Clinical File: Please double check your YEAR_CONTACT "
            "column, it must be an integer in YYYY format <= {year} or "
            "'Unknown', 'Not Collected', 'Not Released', '>89', '<18'.\n"
            "Patient Clinical File: Please double check your INT_CONTACT "
            "column, it must be an integer, '>32485', '<6570', 'Unknown', "
            "'Not Released' or 'Not Collected'.\n"
            "Patient Clinical File: Please double check your INT_DOD column, "
            "it must be an integer, '>32485', '<6570', 'Unknown', "
            "'Not Collected', 'Not Released' or 'Not Applicable'.\n"
            "Patient Clinical File: Please double check your DEAD column, "
            "it must be True, False, 'Unknown', "
            "'Not Released' or 'Not Collected'.\n"
            "Patient Clinical File: Please double check your PRIMARY_RACE "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
            "Patient Clinical File: Please double check your SECONDARY_RACE "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
            "Patient Clinical File: Please double check your TERTIARY_RACE "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
            "Patient Clinical File: Please double check your SEX column.  "
            "This column must only be these values: 1, 2, 99\n"
            "Patient Clinical File: Please double check your ETHNICITY "
            "column.  This column must only be these values: 1, 2, 3, 4, 99\n"
        )
        expectedWarnings = (
            "Sample Clinical File: All patients must have associated sample "
            "information. These patients are missing sample data: ID6\n"
            "Sample Clinical File: Some SAMPLE_IDs have conflicting SEX and "
            "ONCOTREE_CODES: ID2-1,ID5-1\n")
        assert error == expectedErrors.format(
            year=datetime.datetime.utcnow().year)
        assert warning == expectedWarnings


def test_duplicated__validate():
    '''
    Test for duplicated SAMPLE_ID error and and in the case that
    both sample and patient
    are uploaded, it could be a duplicated PATIENT_ID error
    '''
    patientDf = pd.DataFrame(dict(
        PATIENT_ID=["ID1", "ID1", "ID3", "ID4", "ID5"],
        SEX=[1, 2, 1, 2, 99],
        PRIMARY_RACE=[1, 2, 3, 4, 99],
        SECONDARY_RACE=[1, 2, 3, 4, 99],
        TERTIARY_RACE=[1, 2, 3, 4, 99],
        ETHNICITY=[1, 2, 3, 4, 99],
        BIRTH_YEAR=["Unknown", 1990, 1990, 1990, 1990],
        CENTER=["FOO", "FOO", "FOO", "FOO", "FOO"],
        YEAR_DEATH=["Unknown", "Not Collected", "Not Applicable",
                    1990, 1990],
        YEAR_CONTACT=["Unknown", "Not Collected", 1990, 1990, 1990],
        INT_CONTACT=['>32485', '<6570', 'Unknown', 'Not Collected', 2000],
        INT_DOD=['>32485', '<6570', 'Unknown', 'Not Collected',
                 'Not Applicable'],
        DEAD=[True, False, 'Unknown', 'Not Collected', True]))

    sampleDf = pd.DataFrame(dict(
        SAMPLE_ID=["ID1-1", "ID3-1", "ID4-1", "ID5-1"],
        PATIENT_ID=["ID1", "ID3", "ID4", "ID5"],
        AGE_AT_SEQ_REPORT=[100000, 100000, 100000, 100000],
        ONCOTREE_CODE=['AMPCA', 'UNKNOWN', 'AMPCA', 'AMPCA'],
        SAMPLE_TYPE=[1, 3, 4, 4],
        SEQ_ASSAY_ID=['SAGE-1-1', 'SAGE-1', 'SAGE-1', 'SAGE-1'],
        SEQ_DATE=['Jan-2013', 'Jul-2013', 'Oct-2013', 'release']))

    clinicalDf = patientDf.merge(sampleDf, on="PATIENT_ID")
    with mock.patch(
            "genie.process_functions.get_oncotree_code_mappings",
            return_value=onco_map_dict) as mock_get_onco_map:
        error, warning = clin_class._validate(clinicalDf, json_oncotreeurl)
        mock_get_onco_map.called_once_with(json_oncotreeurl)
        expectedErrors = (
            "Sample Clinical File: No duplicated SAMPLE_ID allowed.\n"
            "If there are no duplicated SAMPLE_IDs, and both sample and "
            "patient files are uploaded, then please check to make sure no "
            "duplicated PATIENT_IDs exist in the patient clinical file.\n")

        assert error == expectedErrors
        assert warning == ""


class fake_oncotree():
    text = json.dumps({
        'TISSUE': {
            'children': {
                'AMPCA': {
                    'level': 1,
                    'mainType': 'Ampullary Cancer',
                    'name': 'Ampullary Carcinoma',
                    'children': {
                        'TESTIS': {
                            'level': 2,
                            'mainType': 'Testicular Cancer, NOS',
                            'name': 'Testis',
                            'children': []
                        },
                        'UCEC': {
                            'level': 2,
                            'mainType': 'Endometrial Cancer',
                            'name': 'Endometrial Carcinoma',
                            'children': []
                        }
                    }
                }
            }
        }
    })


expected_onco_mapping = {
    'AMPCA': {
        'CANCER_TYPE': 'Ampullary Cancer',
        'CANCER_TYPE_DETAILED': 'Ampullary Carcinoma',
        'ONCOTREE_PRIMARY_NODE': 'AMPCA',
        'ONCOTREE_SECONDARY_NODE': ''
    },
    'TESTIS': {
        'CANCER_TYPE': 'Testicular Cancer, NOS',
        'CANCER_TYPE_DETAILED': 'Testis',
        'ONCOTREE_PRIMARY_NODE': 'AMPCA',
        'ONCOTREE_SECONDARY_NODE': 'TESTIS'
    },
    'UCEC': {
        'CANCER_TYPE': 'Endometrial Cancer',
        'CANCER_TYPE_DETAILED': 'Endometrial Carcinoma',
        'ONCOTREE_PRIMARY_NODE': 'AMPCA',
        'ONCOTREE_SECONDARY_NODE': 'UCEC'
    }
}


def test_get_oncotree_code_mappings():
    from genie import process_functions
    with mock.patch(
            "genie.process_functions.retry_get_url",
            return_value=fake_oncotree) as retry_get_url:
        onco_mapping = \
            process_functions.get_oncotree_code_mappings(json_oncotreeurl)
        retry_get_url.called_once_with(json_oncotreeurl)
        assert onco_mapping == expected_onco_mapping


def test__check_year_no_errors():
    """Tests perfect checking year validation function"""
    perfectdf = pd.DataFrame(
        {"BIRTH_YEAR": ["Unknown", 1990, 1990, 1990, 1990]}
    )
    error = genie_registry.clinical._check_year(perfectdf, "BIRTH_YEAR", "Filename",
                                       allowed_string_values = ["Unknown"])
    assert error == ''


def test__check_year_too_big_year():
    """Tests year can't be greater than current year"""
    year_now = datetime.datetime.utcnow().year

    errordf = pd.DataFrame(
        {"BIRTH_YEAR": ["Unknown", year_now+1, 1990, 1990, 1990]}
    )
    error = genie_registry.clinical._check_year(
        errordf, "BIRTH_YEAR", "Filename",
        allowed_string_values = ["Unknown"]
    )
    assert error == (
        "Filename: Please double check your BIRTH_YEAR column, "
        f"it must be an integer in YYYY format <= {year_now} or 'Unknown'.\n"
    )


def test__check_year_invalid():
    """Tests year can't be greater than current year"""
    year_now = datetime.datetime.utcnow().year

    errordf = pd.DataFrame(
        {"BIRTH_YEAR": ["Unknown", 1990, 1990, 1990, 1990]}
    )
    error = genie_registry.clinical._check_year(
        errordf, "BIRTH_YEAR", "Filename"
    )
    assert error == (
        "Filename: Please double check your BIRTH_YEAR column, "
        f"it must be an integer in YYYY format <= {year_now}.\n"
    )


def test_remap_clinical_values_sampletype():
    """Test adding of sample type detailed when remapping clinical values"""
    testdf = pd.DataFrame(
        {"SAMPLE_TYPE": [1, 2, 99]}
    )
    expecteddf = pd.DataFrame(
        {"SAMPLE_TYPE": ["Test", "Why", "Unknown"],
         "SAMPLE_TYPE_DETAILED": ["non", "asdf", "asdfasdf"]}
    )
    remappeddf = genie_registry.clinical.remap_clinical_values(
        testdf, sexdf, no_nan, sexdf, no_nan
    )
    assert expecteddf.equals(remappeddf)


@pytest.mark.parametrize("col", ["SEX", "PRIMARY_RACE", "SECONDARY_RACE",
                                 "TERTIARY_RACE", "ETHNICITY"])
def test_remap_clinical_values(col):
    """Test Remapping clinical values"""
    testdf = pd.DataFrame(
        {col: [1, 2, 99]}
    )
    expecteddf = pd.DataFrame(
        {col: ["Male", "Female", "Unknown"]}
    )
    remappeddf = genie_registry.clinical.remap_clinical_values(
        testdf, sexdf, sexdf, sexdf, sexdf
    )
    assert expecteddf.equals(remappeddf)



from io import StringIO
import os
import logging

import pandas as pd

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions

logger = logging.getLogger(__name__)


def _check_tsa1_tsa2(df):
    """If maf file has both TSA1 and TSA2,
    TSA1 must equal REF, or TSA1 must equal TSA2.
    """
    tsa2_col_exist = process_functions.checkColExist(df, "TUMOR_SEQ_ALLELE2")
    tsa1_col_exist = process_functions.checkColExist(df, "TUMOR_SEQ_ALLELE1")
    ref_col_exist = process_functions.checkColExist(df, "REFERENCE_ALLELE")
    error = ""
    if tsa2_col_exist and tsa1_col_exist and ref_col_exist:
        tsa1_eq_ref = all(df['TUMOR_SEQ_ALLELE1'] == df['REFERENCE_ALLELE'])
        tsa1_eq_tsa2 = all(
            df['TUMOR_SEQ_ALLELE1'] == df['TUMOR_SEQ_ALLELE2']
        )
        if not (tsa1_eq_ref or tsa1_eq_tsa2):
            error = (
                "maf: Contains both "
                "TUMOR_SEQ_ALLELE1 and TUMOR_SEQ_ALLELE2 columns. "
                "The values in TUMOR_SEQ_ALLELE1 must be the same as "
                "all the values in REFERENCE_ALELLE OR TUMOR_SEQ_ALLELE2."
            )
    return error



def _check_allele_col(df, col):
    """
    Check the Allele column is correctly formatted.

    Args:
        df: mutation dataframe
        col: Column header name

    Returns:
        error, warning

    """
    col_exist = process_functions.checkColExist(df, col)
    error = ""
    warning = ""
    if col_exist:
        # CHECK: The value "NA" can't be used as a placeholder
        if sum(df[col].fillna('') == "NA") > 0:
            warning = (
                "maf: "
                f"{col} column contains 'NA' values, "
                "which cannot be placeholders for blank values.  "
                "Please put in empty strings for blank values.\n"
            )
        # CHECK: There can't be any null values
        if sum(df[col].isnull()) > 0:
            error = (
                f"maf: {col} can't have any blank or null values.\n"
            )

    return error, warning


class maf(FileTypeFormat):
    '''
    MAF file format validation / processing
    '''
    _fileType = "maf"

    _process_kwargs = []

    def _validateFilename(self, filePath):
        '''
        Validates filename.  Should be
        data_mutations_extended_CENTER.txt
        '''
        assert os.path.basename(filePath[0]) == \
            "data_mutations_extended_{}.txt".format(self.center)

    def process_steps(self, df):
        """The processing of maf files is specific to GENIE, so
        not included in this function"""
        logger.info("Please run with `--process mutation` parameter "
                    "if you want to reannotate the mutation files")
        return None

    def _validate(self, mutationDF):
        """
        This function validates the mutation file to make sure it
        adheres to the mutation SOP.

        Args:
            mutationDF: mutation dataframe

        Returns:
            Text with all the errors in the mutation file
        """

        first_header = ['CHROMOSOME', 'HUGO_SYMBOL', 'TUMOR_SAMPLE_BARCODE']
        SP = self._fileType == "mafSP"
        if SP:
            correct_column_headers = [
                'CHROMOSOME', 'START_POSITION', 'REFERENCE_ALLELE',
                'TUMOR_SAMPLE_BARCODE', 'TUMOR_SEQ_ALLELE2']
            # T_REF_COUNT + T_ALT_COUNT = T_DEPTH
        else:
            correct_column_headers = [
                'CHROMOSOME', 'START_POSITION', 'REFERENCE_ALLELE',
                'TUMOR_SAMPLE_BARCODE', 'T_ALT_COUNT', 'TUMOR_SEQ_ALLELE2']
            # T_REF_COUNT + T_ALT_COUNT = T_DEPTH
        optional_headers = \
            ['T_REF_COUNT', 'N_DEPTH', 'N_REF_COUNT', 'N_ALT_COUNT']

        mutationDF.columns = [col.upper() for col in mutationDF.columns]

        # total_error = ""
        total_error = StringIO()
        warning = StringIO()

        # CHECK: Everything in correct_column_headers must be in mutation file
        if not all([process_functions.checkColExist(mutationDF, i)
                    for i in correct_column_headers]):
            total_error.write(
                "maf: Must at least have these headers: {}. "
                "If you are writing your maf file with R, please make"
                "sure to specify the 'quote=FALSE' parameter.\n".format(
                    ",".join([i for i in correct_column_headers
                              if i not in mutationDF.columns.values])))
        else:
            # CHECK: First column must be in the first_header list
            if mutationDF.columns[0] not in first_header:
                total_error.write(
                    "maf: First column header must be "
                    "one of these: {}.\n".format(", ".join(first_header))
                )
            # No duplicated values
            primary_cols = ['CHROMOSOME', 'START_POSITION',
                            'REFERENCE_ALLELE', 'TUMOR_SAMPLE_BARCODE',
                            'TUMOR_SEQ_ALLELE2']
            # Strip white space if string column
            for col in primary_cols:
                if mutationDF[col].dtype == object:
                    mutationDF[col] = mutationDF[col].str.strip()
            duplicated_idx = mutationDF.duplicated(primary_cols)
            # Find samples with duplicated variants
            duplicated_variants = mutationDF['TUMOR_SAMPLE_BARCODE'][
                duplicated_idx
            ].unique().tolist()

            if duplicated_idx.any():
                total_error.write(
                    "maf: Must not have duplicated variants. "
                    "Samples with duplicated variants: "
                    f"{', '.join(duplicated_variants)}\n"
                )

        t_depth_exists = process_functions.checkColExist(mutationDF,
                                                         "T_DEPTH")
        t_ref_exists = process_functions.checkColExist(mutationDF,
                                                       "T_REF_COUNT")
        if not t_depth_exists and not t_ref_exists and not SP:
            total_error.write(
                "maf: If missing T_DEPTH, must have T_REF_COUNT!\n"
            )
        numerical_cols = ['T_DEPTH', 'T_ALT_COUNT', 'T_REF_COUNT',
                          'N_DEPTH', 'N_REF_COUNT', 'N_ALT_COUNT']
        for col in numerical_cols:
            col_exists = process_functions.checkColExist(mutationDF, col)
            if col_exists:
                # Since NA is an allowed value, when reading in the dataframe
                # the 'NA' string is not converted.  This will convert all
                # 'NA' values in the numerical columns into actual float('nan')
                mutationDF.loc[mutationDF[col] == "NA", col] = float('nan')
                # Attempt to convert column to float
                try:
                    mutationDF[col] = mutationDF[col].astype(float)
                except ValueError:
                    pass
                if mutationDF[col].dtype not in [int, float]:
                    total_error.write(
                        f"maf: {col} must be a numerical column.\n"
                    )

        # CHECK: Must have TUMOR_SEQ_ALLELE2
        error, warn = _check_allele_col(mutationDF, "TUMOR_SEQ_ALLELE2")
        total_error.write(error)
        warning.write(warn)

        # CHECK: Mutation file would benefit from columns in optional_headers
        if not all([process_functions.checkColExist(mutationDF, i)
                    for i in optional_headers]) and not SP:
            warning.write(
                "maf: Does not have the column headers that can give extra "
                "information to the processed maf: {}.\n".format(
                    ", ".join([
                        i for i in optional_headers
                        if i not in mutationDF.columns.values]))
            )

        # CHECK: Must have REFERENCE_ALLELE
        error, warn = _check_allele_col(mutationDF, "REFERENCE_ALLELE")
        total_error.write(error)
        warning.write(warn)

        if process_functions.checkColExist(mutationDF, "CHROMOSOME"):
            # CHECK: Chromosome column can't have any values that start
            # with chr or have any WT values
            invalid_values = [
                str(i).startswith("chr") or str(i) == "WT"
                for i in mutationDF['CHROMOSOME']
            ]
            if sum(invalid_values) > 0:
                total_error.write(
                    "maf: CHROMOSOME column cannot have any values that "
                    "start with 'chr' or any 'WT' values.\n"
                )

        error = _check_tsa1_tsa2(mutationDF)
        total_error.write(error)

        return total_error.getvalue(), warning.getvalue()

    def _get_dataframe(self, filePathList):
        """Get mutation dataframe"""
        # Must do this because pandas.read_csv will allow for a file to
        # have more column headers than content.  E.g.
        # A,B,C,D,E
        # 1,2
        # 2,3
        with open(filePathList[0], 'r') as maf_f:
            firstline = maf_f.readline()
            if firstline.startswith("#"):
                firstline = maf_f.readline()
            secondline = maf_f.readline()

        if len(firstline.split("\t")) != len(secondline.split("\t")):
            raise ValueError("Number of fields in a line do not match the "
                             "expected number of columns")

        mutationdf = pd.read_csv(filePathList[0],
                                 sep="\t",
                                 comment="#",
                                 # Keep the value 'NA'
                                 na_values=['-1.#IND', '1.#QNAN', '1.#IND',
                                            '-1.#QNAN', '#N/A N/A', 'NaN',
                                            '#N/A', 'N/A', '#NA', 'NULL',
                                            '-NaN', 'nan', '-nan', ''],
                                 keep_default_na=False,
                                 # This is to check if people write files
                                 # with R, quote=T
                                 quoting=3,
                                 # Retain completely blank lines so that
                                 # validator will cause the file to fail
                                 skip_blank_lines=False)
        return mutationdf



"""Process mutation files"""
import logging
import os
import shutil
import subprocess
import tempfile

import pandas as pd
import synapseclient
from synapseclient import Synapse
from synapseclient.core.exceptions import SynapseTimeoutError

from . import process_functions

logger = logging.getLogger(__name__)

# Some columns are already capitalized, so they aren't included here
MAF_COL_MAPPING = {
    'HUGO_SYMBOL': 'Hugo_Symbol',
    'ENTREZ_GENE_ID': 'Entrez_Gene_Id',
    'CENTER': 'Center',
    'NCBI_BUILD': 'NCBI_Build',
    'CHROMOSOME': 'Chromosome',
    'START_POSITION': 'Start_Position',
    'END_POSITION': 'End_Position',
    'STRAND': 'Strand',
    'VARIANT_CLASSIFICATION': 'Variant_Classification',
    'VARIANT_TYPE': 'Variant_Type',
    'REFERENCE_ALLELE': 'Reference_Allele',
    'TUMOR_SEQ_ALLELE1': 'Tumor_Seq_Allele1',
    'TUMOR_SEQ_ALLELE2': 'Tumor_Seq_Allele2',
    'DBSNP_RS': 'dbSNP_RS',
    'DBSNP_VAL_STATUS': 'dbSNP_Val_Status',
    'TUMOR_SAMPLE_BARCODE': 'Tumor_Sample_Barcode',
    'MATCHED_NORM_SAMPLE_BARCODE': 'Matched_Norm_Sample_Barcode',
    'MATCH_NORM_SEQ_ALLELE1': 'Match_Norm_Seq_Allele1',
    'MATCH_NORM_SEQ_ALLELE2': 'Match_Norm_Seq_Allele2',
    'TUMOR_VALIDATION_ALLELE1': 'Tumor_Validation_Allele1',
    'TUMOR_VALIDATION_ALLELE2': 'Tumor_Validation_Allele2',
    'MATCH_NORM_VALIDATION_ALLELE1': 'Match_Norm_Validation_Allele1',
    'MATCH_NORM_VALIDATION_ALLELE2': 'Match_Norm_Validation_Allele2',
    'VERIFICATION_STATUS': 'Verification_Status',
    'VALIDATION_STATUS': 'Validation_Status',
    'MUTATION_STATUS': 'Mutation_Status',
    'SEQUENCING_PHASE': 'Sequencing_Phase',
    'SEQUENCE_SOURCE': 'Sequence_Source',
    'VALIDATION_METHOD': 'Validation_Method',
    'SCORE': 'Score',
    'BAM_FILE': 'BAM_File',
    'SEQUENCER': 'Sequencer',
    'T_REF_COUNT': 't_ref_count',
    'T_ALT_COUNT': 't_alt_count',
    'N_REF_COUNT': 'n_ref_count',
    'N_ALT_COUNT': 'n_alt_count',
    'ALLELE': 'Allele',
    'AMINO_ACID_CHANGE': 'amino_acid_change',
    'AMINO_ACIDS': 'Amino_acids',
    'CDS_POSITION': 'CDS_position',
    'CODONS': 'Codons',
    'CONSEQUENCE': 'Consequence',
    'EXISTING_VARIATION': 'Existing_variation',
    'EXON_NUMBER': 'Exon_Number',
    'FEATURE': 'Feature',
    'FEATURE_TYPE': 'Feature_type',
    'GENE': 'Gene',
    'HGVSC': 'HGVSc',
    'HGVSP': 'HGVSp',
    'HGVSP_SHORT': 'HGVSp_Short',
    'HOTSPOT': 'Hotspot',
    'MA:FIMPACT': 'MA:FImpact',
    'MA:LINK.MSA': 'MA:link.MSA',
    'MA:LINK.PDB': 'MA:link.PDB',
    'MA:LINK.VAR': 'MA:link.var',
    'MA:PROTEIN.CHANGE': 'MA:protein.change',
    'POLYPHEN': 'PolyPhen',
    'PROTEIN_POSITION': 'Protein_position',
    'REFSEQ': 'RefSeq',
    'TRANSCRIPT': 'transcript',
    'TRANSCRIPT_ID': 'Transcript_ID',
    'ALL_EFFECTS': 'all_effects',
    'CDNA_CHANGE': 'cdna_change',
    'CDNA_POSITION': 'cDNA_position',
    'N_DEPTH': 'n_depth',
    'T_DEPTH': 't_depth'
}

KNOWN_STRING_COLS = ['IS_NEW', 'ALLELE_NUM', 'Chromosome', 'CLIN_SIG',
                     'MOTIF_NAME', 'HIGH_INF_POS', 'MINIMISED',
                     'CHROMOSOME', 'VERIFICATION_STATUS',
                     'VALIDATION_STATUS', 'MUTATION_STATUS',
                     'SEQUENCE_SOURCE', 'SEQUENCER', 'REPORT_AF',
                     'CDNA_CHANGE', 'AMINO_ACID_CHANGE',
                     'TRANSCRIPT', 'STRAND_VEP',
                     'HGNC_ID', 'PUBMED', 'PICK']


def _rename_column_headers(dataframe: pd.DataFrame,
                           col_map: dict) -> pd.DataFrame:
    """Rename dataframe column headers

    Args:
        dataframe: pandas dataframe
        col_map: Column Mapping {column_name: new_column_name,}

    Returns:
        Dataframe with new columns

    """
    dataframe = dataframe.rename(columns=col_map)
    return dataframe


def _convert_to_str_dtype(column_types, known_string_cols):
    """Sometimes the deteremined dtype is incorrect based off the first
    100 rows, update the incorrect dtypes.
    """
    for str_col in known_string_cols:
        if column_types.get(str_col):
            column_types[str_col] = 'object'
    return column_types


def determine_dtype(path: str):
    """Reads in a dataframe partially and determines the dtype of columns"""
    subset_df = pd.read_csv(path, nrows=100, sep="\t")
    dtypes = subset_df.dtypes
    colnames = dtypes.index
    types = [i.name for i in dtypes.values]
    column_types = dict(zip(colnames, types))
    return column_types


def move_maf(mutation_path, input_files_dir):
    """Moves maf files into processing directory. Maf file's column headers
    are renamed if necessary"""
    header_df = pd.read_csv(mutation_path, sep="\t",
                            index_col=0, nrows=1, comment="#")
    # If any column headers need to be remapped, remap
    if sum(header_df.columns.isin(MAF_COL_MAPPING.keys())) > 0:
        filename = os.path.basename(mutation_path)
        column_types = determine_dtype(mutation_path)
        new_column_types = _convert_to_str_dtype(column_types,
                                                 KNOWN_STRING_COLS)
        mafdf = pd.read_csv(mutation_path, sep="\t", dtype=new_column_types)
        mafdf = _rename_column_headers(mafdf, col_map=MAF_COL_MAPPING)
        mafdf.to_csv(os.path.join(input_files_dir, filename),
                     sep="\t", index=False)
    else:
        shutil.copy(mutation_path, input_files_dir)


def move_mutation(mutation_path, input_files_dir):
    """Move mutation file into processing directory"""
    # If mutation file is vcf, just copy
    if mutation_path.endswith(".vcf"):
        shutil.copy(mutation_path, input_files_dir)
    else:
        move_maf(mutation_path, input_files_dir)


def process_mutation_workflow(syn: Synapse, center: str,
                              validfiles: pd.DataFrame,
                              genie_annotation_pkg: str,
                              database_mappingdf: str,
                              workdir: str) -> str:
    """Process vcf/maf workflow

    Args:
        syn: Synapse connection
        center: Center name
        validfiles: Center validated files
        genie_annotation_pkg: Genome Nexus annotation tools
        database_mappingdf: Database to synapse id mapping dataframe
        workdir: Working directory

    Returns:
        Annotated Maf Path

    """
    # Get valid files
    mutation_files = validfiles['fileType'].isin(["maf", "vcf"])
    valid_mutation_files = validfiles['path'][mutation_files].tolist()
    # If there are no valid mutation files, return
    if not valid_mutation_files:
        logger.info("No mutation data")
        return
    # Certificate to use GENIE Genome Nexus
    syn.get("syn22053204", ifcollision="overwrite.local",
            downloadLocation=genie_annotation_pkg)
    # Genome Nexus Jar file
    syn.get("syn22084320", ifcollision="overwrite.local",
            downloadLocation=genie_annotation_pkg)

    annotated_maf_path = annotate_mutation(
        center=center,
        mutation_files=valid_mutation_files,
        genie_annotation_pkg=genie_annotation_pkg,
        workdir=workdir
    )

    maf_tableid = database_mappingdf.Id[
        database_mappingdf['Database'] == 'vcf2maf'
    ].iloc[0]
    flatfiles_synid = database_mappingdf.Id[
        database_mappingdf['Database'] == "centerMaf"
    ].iloc[0]
    # Split into narrow maf and store into db / flat file
    split_and_store_maf(syn=syn,
                        center=center,
                        maf_tableid=maf_tableid,
                        annotated_maf_path=annotated_maf_path,
                        flatfiles_synid=flatfiles_synid,
                        workdir=workdir)

    return annotated_maf_path


def annotate_mutation(center: str, mutation_files: list,
                      genie_annotation_pkg: str,
                      workdir: str) -> str:
    """Process vcf/maf files

    Args:
        center: Center name
        mutation_files: list of mutation files
        genie_annotation_pkg: Path to GENIE annotation package

    Returns:
        Path to final maf
    """
    input_files_dir = tempfile.mkdtemp(dir=workdir)
    output_files_dir = tempfile.mkdtemp(dir=workdir)

    for mutation_file in mutation_files:
        move_mutation(mutation_file, input_files_dir)

    merged_maf_path = os.path.join(
        output_files_dir, f"data_mutations_extended_{center}.txt"
    )
    annotater_cmd = ['bash', os.path.join(genie_annotation_pkg,
                                          'annotation_suite_wrapper.sh'),
                     f'-i={input_files_dir}',
                     f'-o={output_files_dir}',
                     f'-m={merged_maf_path}',
                     f'-c={center}',
                     '-s=WXS',
                     f'-p={genie_annotation_pkg}']

    subprocess.check_call(annotater_cmd)

    return merged_maf_path


def append_or_createdf(dataframe: pd.DataFrame, filepath: str):
    """Creates a file with the dataframe or appends to a existing file.

    Args:
        df: pandas.dataframe to write out
        filepath: Filepath to append or create

    """
    if not os.path.exists(filepath) or os.stat(filepath).st_size == 0:
        dataframe.to_csv(filepath, sep="\t", index=False)
    else:
        dataframe.to_csv(filepath, sep="\t", mode='a', index=False,
                         header=None)


def store_full_maf(syn: Synapse, filepath: str, parentid: str):
    """Stores full maf file

    Args:
        syn: Synapse connection
        filepath: Path to file
        parentid: Synapse container id

    """
    syn.store(synapseclient.File(filepath, parentId=parentid))


def store_narrow_maf(syn: Synapse, filepath: str, maf_tableid: str):
    """
    Stores the narrow maf in Synapse Table

    Args:
        syn: Synapse connection
        filepath: Path to maf file
        maf_tableid: database synid

    """
    logger.info(f'STORING {filepath}')
    # database = syn.get(maf_tableid)
    try:
        update_table = synapseclient.Table(maf_tableid, filepath,
                                           separator="\t")
        syn.store(update_table)
    except SynapseTimeoutError:
        # This error occurs because of waiting for table to index.
        # Don't worry about this.
        pass


def format_maf(mafdf: pd.DataFrame, center: str) -> pd.DataFrame:
    """Format maf file, shortens the maf file length

    Args:
        mafdf: mutation dataframe
        center: Center name

    Returns:
        Formatted mutation dataframe"""
    mafdf['Center'] = center
    mafdf['Tumor_Sample_Barcode'] = [
        process_functions.checkGenieId(i, center)
        for i in mafdf['Tumor_Sample_Barcode']
    ]

    mafdf['Sequence_Source'] = float('nan')
    mafdf['Sequencer'] = float('nan')
    mafdf['Validation_Status'][
        mafdf['Validation_Status'].isin(["Unknown", "unknown"])
    ] = ''

    return mafdf


def split_and_store_maf(syn: 'Synapse', center: str, maf_tableid: str,
                        annotated_maf_path: str, flatfiles_synid: str,
                        workdir: str):
    """Separates annotated maf file into narrow and full maf and stores them

    Args:
        syn: Synapse connection
        center: Center
        maf_tableid: Mutation table synapse id
        annotated_maf_path: Annotated maf
        flatfiles_synid: GENIE flat files folder

    """
    narrow_maf_cols = [col['name']
                       for col in syn.getTableColumns(maf_tableid)
                       if col['name'] != 'inBED']
    full_maf_path = os.path.join(
        workdir, center, "staging",
        f"data_mutations_extended_{center}.txt"
    )
    narrow_maf_path = os.path.join(
        workdir, center, "staging",
        f"data_mutations_extended_{center}_MAF_narrow.txt"
    )
    maf_chunks = pd.read_csv(annotated_maf_path, sep="\t", chunksize=100000)

    for maf_chunk in maf_chunks:
        maf_chunk = format_maf(maf_chunk, center)
        append_or_createdf(maf_chunk, full_maf_path)
        narrow_maf_chunk = maf_chunk[narrow_maf_cols]
        append_or_createdf(narrow_maf_chunk, narrow_maf_path)

    store_narrow_maf(syn, narrow_maf_path, maf_tableid)
    # Store MAF flat file into synapse
    store_full_maf(syn, full_maf_path, flatfiles_synid)



"""Test assay information validation and processing"""
import copy
from unittest.mock import patch, create_autospec

import pandas as pd
import pytest
import synapseclient

from genie_registry.assay import Assayinfo
from genie import process_functions

GDC_DATA_DICT = {
    'properties': {
        'library_strategy': {
            'enum': ['value1', 'value2']
        },
        'library_selection': {
            'enum': ['value1', 'value2']
        },
        'platform': {
            'enum': ['value1', 'value2']
        },
        'instrument_model': {
            'enum': ['value1', 'value2']
        },
        'target_capture_kit': {
            'enum': ['value1', 'value2']
        }
    }
}

SYN = create_autospec(synapseclient.Synapse)
ASSAY_INFO = Assayinfo(SYN, "SAGE")


def test_filetype():
    assert ASSAY_INFO._fileType == "assayinfo"


def test_invalidname__validatefilename():
    """Assertion error thrown for wrong filename"""
    with pytest.raises(AssertionError):
        ASSAY_INFO.validateFilename(['foo'])


def test_correct__validatefilename():
    """Filetype returned when filename valid"""
    assert ASSAY_INFO.validateFilename(
        ["assay_information.yaml"]) == "assayinfo"


def test_validinput__validate():
    """Valid input should have no errors or warnings"""
    assay_info_dict = {
        'SEQ_ASSAY_ID': ['SAGE-1', 'SAGE-3'],
        'is_paired_end': [True, False],
        'library_strategy': ['value1', 'value2'],
        'library_selection': ['value1', 'value2'],
        'platform': ['value1', 'value2'],
        'instrument_model': ['value1', 'value2'],
        'target_capture_kit': ['value1', 'value2'],
        'variant_classifications': ['Frame_Shift_Ins', 'Frame_Shift_Ins'],
        'read_length': [22, float('nan')],
        'number_of_genes': [5, 20],
        'gene_padding': [10, None],
        'calling_strategy': ['tumor_only', 'tumor_normal'],
        'specimen_tumor_cellularity': ['>10%', '>20%'],
        'alteration_types': ['snv;small_indels', 'intragenic_cna'],
        'preservation_technique': ['FFPE', 'FFPE;fresh_frozen'],
        'coverage': ['hotspot_regions;introns', 'introns']}

    assay_info_df = pd.DataFrame(assay_info_dict)
    test_dict = copy.deepcopy(GDC_DATA_DICT)
    with patch.object(process_functions, "get_gdc_data_dictionary",
                      return_value=test_dict) as patch_get_gdc:
        error, warning = ASSAY_INFO._validate(assay_info_df)
        assert error == ''
        assert warning == ''


def test__missingcols__validate():
    """Test missing columns"""
    assay_info_df = pd.DataFrame()
    test_dict = copy.deepcopy(GDC_DATA_DICT)
    with patch.object(process_functions, "get_gdc_data_dictionary",
                      return_value=test_dict) as patch_get_gdc:
        error, warning = ASSAY_INFO._validate(assay_info_df)
    expected_errors = (
        'Assay_information.yaml: Must have SEQ_ASSAY_ID column.\n'
        'Assay_information.yaml: Must have is_paired_end column.\n'
        'Assay_information.yaml: Must have library_selection column.\n'
        'Assay_information.yaml: Must have library_strategy column.\n'
        'Assay_information.yaml: Must have platform column.\n'
        'Assay_information.yaml: Must have instrument_model column.\n'
        'Assay_information.yaml: Must have target_capture_kit column.\n'
        'Assay_information.yaml: Must have read_length column.\n'
        'Assay_information.yaml: Must have number_of_genes column.\n'
        'Assay_information.yaml: Must have calling_strategy column.\n'
        'Assay_information.yaml: '
        'Must have specimen_tumor_cellularity column.\n'
        'Assay_information.yaml: Must have alteration_types column.\n'
        'Assay_information.yaml: Must have preservation_technique column.\n'
        'Assay_information.yaml: Must have coverage column.\n')
    assert error == expected_errors

    expected_warnings = (
        "Assay_information.yaml: Doesn't have variant_classifications column. "
        "This column will be added\n"
        "Assay_information.yaml: gene_padding is "
        "by default 10 if not specified.\n")
    assert warning == expected_warnings


def test_fillcols__process():
    '''
    Standardization of SEQ_ASSAY_ID
    Add in CENTER, gene_padding, and variant_classifications if missing
    '''

    assay_info_dict = {'SEQ_ASSAY_ID': ['SAGE-Foo_1'],
                       'SEQ_PIPELINE_ID': ['SAGE_Foo']}
    assay_info_df = pd.DataFrame(assay_info_dict)
    processed_assay_df = ASSAY_INFO._process(assay_info_df)
    expected_assay_df = pd.DataFrame(
        {'SEQ_ASSAY_ID': ['SAGE-FOO-1'],
         'SEQ_PIPELINE_ID': ['SAGE-FOO'],
         'gene_padding': [10],
         'variant_classifications': [float('nan')],
         'CENTER': ['SAGE']})

    assert expected_assay_df.equals(
        processed_assay_df[expected_assay_df.columns])


def test_default10__process():
    '''
    gene_padding default 10
    '''

    assay_info_dict = {'SEQ_ASSAY_ID': ['SAGE-1', 'SAGE-2'],
                       'SEQ_PIPELINE_ID': ['SAGE-1', 'SAGE-2'],
                       'gene_padding': [20, float('nan')],
                       'variant_classifications': ['test', 'test']}
    assay_info_df = pd.DataFrame(assay_info_dict)
    processed_assay_df = ASSAY_INFO._process(assay_info_df)
    expected_assay_df = pd.DataFrame(
        {'SEQ_ASSAY_ID': ['SAGE-1', 'SAGE-2'],
         'SEQ_PIPELINE_ID': ['SAGE-1', 'SAGE-2'],
         'gene_padding': [20, 10],
         'variant_classifications': ['test', 'test'],
         'CENTER': ['SAGE', 'SAGE']})
    assert expected_assay_df.equals(
        processed_assay_df[expected_assay_df.columns])


def test_invalid__validate():
    assay_info_dict = {
        'SEQ_ASSAY_ID': ['SAGE-1', 'SAGE-3'],
        'is_paired_end': [True, "foo"],
        'library_strategy': ['foo', 'ChIP-Seq'],
        'library_selection': ['foo', 'PCR'],
        'platform': ['foo', 'Illumina'],
        'instrument_model': ['foo', 'Illumina HiSeq 4000'],
        'variant_classifications': ['foo', 'Frame_Shift_Ins'],
        'target_capture_kit': ['foo', 'doo'],
        'read_length': [22, 'foo'],
        'number_of_genes': [5, 'foo'],
        'gene_padding': [10, 'foo'],
        'calling_strategy': ['tumor_ony', 'tumor_normal'],
        'specimen_tumor_cellularity': ['>10', '>20%'],
        'alteration_types': ['snv;small_indel', 'intragenic_cna'],
        'preservation_technique': ['FPE', 'FFPE;fresh_frozen'],
        'coverage': ['hotsot_regions;introns', 'introns']}

    assay_info_df = pd.DataFrame(assay_info_dict)
    # Must use deepcopy because a dict.copy is a shallow copy
    # Which just points to reference keys
    test_dict = copy.deepcopy(GDC_DATA_DICT)
    with patch.object(process_functions, "get_gdc_data_dictionary",
                      return_value=test_dict) as patch_get_gdc:
        error, warning = ASSAY_INFO._validate(assay_info_df)
        expected_errors = (
            "Assay_information.yaml: "
            "Please double check your is_paired_end column.  "
            "This column must only be these values: True, False\n"
            "Assay_information.yaml: "
            "Please double check your library_selection column.  "
            "This column must only be these values: value1, value2\n"
            "Assay_information.yaml: "
            "Please double check your library_strategy column.  "
            "This column must only be these values: value1, value2\n"
            "Assay_information.yaml: "
            "Please double check your platform column.  "
            "This column must only be these values: value1, value2\n"
            "Assay_information.yaml: "
            "Please double check your instrument_model column.  "
            "This column must only be these values: value1, value2, None\n"
            "Assay_information.yaml: "
            "Please double check your variant_classifications column.  "
            "This column must only be these values: Splice_Site, "
            "Nonsense_Mutation, Frame_Shift_Del, Frame_Shift_Ins, "
            "Nonstop_Mutation, Translation_Start_Site, In_Frame_Ins, "
            "In_Frame_Del, Missense_Mutation, Intron, Splice_Region, "
            "Silent, RNA, 5'UTR, 3'UTR, IGR, 5'Flank, 3'Flank, None\n"
            "Assay_information.yaml: "
            "Please double check your read_length.  "
            "It must be an integer or null.\n"
            "Assay_information.yaml: "
            "Please double check your number_of_genes. "
            "It must be an integer.\n"
            "Assay_information.yaml: "
            "Please double check your gene_padding. "
            "It must be an integer or blank.\n"
            "Assay_information.yaml: "
            "Please double check your calling_strategy column.  This "
            "column must only be these values: tumor_only, tumor_normal\n"
            "Assay_information.yaml: "
            "Please double check your specimen_tumor_cellularity. "
            "It must in this format >(num)%. ie. >10%\n"
            "Assay_information.yaml: "
            "Please double check your alteration_types column.  "
            "This column must only be these values: snv, small_indels, "
            "gene_level_cna, intragenic_cna, structural_variants\n"
            "Assay_information.yaml: "
            "Please double check your preservation_technique column.  "
            "This column must only be these values: FFPE, fresh_frozen\n"
            "Assay_information.yaml: "
            "Please double check your coverage column.  "
            "This column must only be these values: "
            "hotspot_regions, coding_exons, introns, promoters\n")

        patch_get_gdc.called_once_with("read_group")
        assert error == expected_errors
        assert warning == ''



from datetime import datetime
import os
from unittest import mock
from unittest.mock import Mock, patch

import pandas as pd
import pytest
import synapseclient
import synapseutils

from genie import input_to_database, process_functions, process_mutation
from genie_registry.clinical import clinical
import genie.config
from genie.validate import GenieValidationHelper


syn = mock.create_autospec(synapseclient.Synapse)
sample_clinical_synid = 'syn2222'

sample_clinical_entity = synapseclient.File(path='data_clinical_supp_sample_SAGE.txt',
                                            id=sample_clinical_synid,
                                            parentId='syn45678',
                                            name='data_clinical_supp_sample_SAGE.txt',
                                            modifiedOn = '2019-03-24T12:00:00.Z',
                                            md5='44444')

patient_clinical_synid = 'syn11111'
patient_clinical_entity = synapseclient.File(path='data_clinical_supp_patient_SAGE.txt',
                                             id=patient_clinical_synid,
                                             parentId='syn45678',
                                             name='data_clinical_supp_patient_SAGE.txt')

vcf1synid = 'syn6666'
vcf1_entity = synapseclient.File(path='GENIE-SAGE-1-1.vcf',
                                 id=vcf1synid,
                                 parentId='syn45678',
                                 name='GENIE-SAGE-1-1.vcf')
vcf2synid = 'syn8888'
vcf2_entity = synapseclient.File(path='GENIE-SAGE-2-1.vcf',
                                 id=vcf2synid,
                                 parentId='syn45678',
                                 name='GENIE-SAGE-2-1.vcf')
first = (
    [('inputs', "syn12345")],
    [('vcfs', 'syn33333')],
    [('data_clinical_supp_sample_SAGE.txt', sample_clinical_synid),
     ('data_clinical_supp_patient_SAGE.txt', patient_clinical_synid)])
second = (
    [('vcfs', "syn33333")],
    [],
    [('GENIE-SAGE-000-1111.vcf', vcf1synid),
     ('GENIE-SAGE-111-2222.vcf', vcf2synid)])
center = "SAGE"
oncotree_link = "http://oncotree.mskcc.org/api/tumorTypes/tree?version=oncotree_2017_06_21"
center_input_synid = "syn9999"
center_staging_synid = "syn9999"
center_mapping = {'inputSynId': [center_input_synid],
                  'stagingSynId': [center_input_synid],
                  'center': [center]}
center_mapping_df = pd.DataFrame(center_mapping)
validation_statusdf = pd.DataFrame({
    'id': ['syn1234', 'syn2345'],
    'status': ['VALID', 'INVALID'],
    'md5': ['3333', '44444'],
    'name': ['first.txt', 'second.txt'],
    'fileType': ['filetype1', 'filetype2']})
error_trackerdf = pd.DataFrame({
    'id': ['syn2345'],
    'errors': ['Invalid file format'],
    'fileType': ['filetype1']})
emptydf = pd.DataFrame(columns=['id'], dtype=str)

class mock_csv_query_result(object):
    def __init__(self, df):
        self.df = df
    def asDataFrame(self):
        return self.df

# def test_samename_rename_file():
#     '''Test that the file path is not renamed.
#     '''
#     filename = synapseclient.utils.make_bogus_data_file()
#     entity = synapseclient.File(path=filename,
#                                 id='syn012345',
#                                 parentId='syn45678',
#                                 name=os.path.basename(filename))
#     expectedpath = filename
#     new_entity = input_to_database.rename_file(entity)
#     assert new_entity.annotations.expectedPath == expectedpath
#     os.remove(filename)


# def test_diffname_rename_file():
#     '''Test that the file path is renamed.
#     '''
#     filename = synapseclient.utils.make_bogus_data_file()
#     entity = synapseclient.File(path=filename,
#                                 id='syn012345',
#                                 parentId='syn45678',
#                                 name='testname')

#     expectedpath = os.path.join(os.path.dirname(filename), "testname")
#     new_entity = input_to_database.rename_file(entity)
#     assert new_entity.annotations.expectedPath == expectedpath
#     os.remove(filename)


def walk_return():
    '''
    Generator returned by synapseutils.walk
    '''
    yield first
    yield second


def walk_return_empty():
    '''
    Generator returned by synapseutils.walk
    '''
    yield ([], [], [])


def test_main_get_center_input_files():
    '''
    Test to make sure center input files are gotten
    excluding the vcf files since process main is specified
    '''
    syn_get_effects = [sample_clinical_entity, patient_clinical_entity]
    expected_center_file_list = [syn_get_effects]

    calls = [mock.call(sample_clinical_synid, downloadFile=True),
             mock.call(patient_clinical_synid, downloadFile=True)]

    with patch.object(synapseutils, "walk",
                      return_value=walk_return()) as patch_synapseutils_walk,\
         patch.object(syn, "get",
                      side_effect=syn_get_effects) as patch_syn_get:
        center_file_list = input_to_database.get_center_input_files(syn,
                                                                    "syn12345",
                                                                    center)

        assert len(center_file_list) == len(expected_center_file_list)
        assert len(center_file_list[0]) == 2
        assert center_file_list == expected_center_file_list
        patch_synapseutils_walk.assert_called_once_with(syn, 'syn12345')
        patch_syn_get.assert_has_calls(calls)


def test_mutation_get_center_input_files():
    """
    Test to make sure center input files are gotten
    including the vcf files since process vcf is specified
    """
    syn_get_effects = [sample_clinical_entity, patient_clinical_entity,
                       vcf1_entity, vcf2_entity]
    expected_center_file_list = [
        [vcf1_entity], [vcf2_entity],
        [sample_clinical_entity, patient_clinical_entity]]
    calls = [
        mock.call(sample_clinical_synid, downloadFile=True),
        mock.call(patient_clinical_synid, downloadFile=True),
        mock.call(vcf1synid, downloadFile=True),
        mock.call(vcf2synid, downloadFile=True)]

    with patch.object(synapseutils, "walk",
                      return_value=walk_return()) as patch_synapseutils_walk,\
         patch.object(syn, "get",
                      side_effect=syn_get_effects) as patch_syn_get:
        center_file_list = input_to_database.get_center_input_files(
            syn, "syn12345", center, process="mutation"
        )
        assert len(center_file_list) == len(expected_center_file_list)
        assert len(center_file_list[2]) == 2
        assert center_file_list == expected_center_file_list
        patch_synapseutils_walk.assert_called_once_with(syn, 'syn12345')
        patch_syn_get.assert_has_calls(calls)


def test_empty_get_center_input_files():
    '''
    Test that center input files is empty if directory
    pass in is empty
    '''
    with patch.object(synapseutils, "walk",
                      return_value=walk_return_empty()) as patch_synapseutils_walk:
        center_file_list = input_to_database.get_center_input_files(
            syn, "syn12345", center, process="vcf")
        assert center_file_list == []
        patch_synapseutils_walk.assert_called_once_with(syn, 'syn12345')


# @pytest.fixture(params=[
#     # tuple with (input, expectedOutput)
#     (["data_CNA_SAGE.txt"], "cna"),
#     (["data_clinical_supp_SAGE.txt"], "clinical"),
#     (["data_clinical_supp_sample_SAGE.txt",
#       "data_clinical_supp_patient_SAGE.txt"], "clinical")])
# def filename_fileformat_map(request):
#     return request.param


# def test_perfect_get_filetype(filename_fileformat_map):
#     (filepath_list, fileformat) = filename_fileformat_map
#     assert input_to_database.get_filetype(
#         syn, filepath_list, center) == fileformat


# def test_wrongfilename_get_filetype():
#     assert input_to_database.get_filetype(syn, ['wrong.txt'], center) is None


def test_unvalidatedinput_check_existing_file_status():
    '''
    Test the values returned by input that hasn't be validated
    '''
    entity = synapseclient.Entity(id='syn1234')
    entities = [entity]

    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(emptydf), mock_csv_query_result(emptydf), entities)
    assert file_status['to_validate']
    assert file_status['status_list'] == []
    assert file_status['error_list'] == []


def test_valid_check_existing_file_status():
    '''
    Test the values returned by input that is already valid
    '''
    entity = synapseclient.Entity(name='first.txt', id='syn1234', md5='3333')
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf), mock_csv_query_result(error_trackerdf), entities)
    assert not file_status['to_validate']
    assert file_status['status_list'] == ['VALID']
    assert file_status['error_list'] == []


def test_invalid_check_existing_file_status():
    '''
    Test the values returned by input that is invalid
    '''
    entity = synapseclient.Entity(name='second.txt', id='syn2345', md5='44444')
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf), mock_csv_query_result(error_trackerdf), entities)
    assert not file_status['to_validate']
    assert file_status['status_list'] == ['INVALID']
    assert file_status['error_list'] == ['Invalid file format']


def test_nostorederrors_check_existing_file_status():
    '''
    If there is no error uploaded, must re-validate file
    '''
    entity = synapseclient.Entity(name='second.txt', id='syn2345', md5='44444')
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf), mock_csv_query_result(emptydf), entities)
    assert file_status['to_validate']
    assert file_status['status_list'] == ['INVALID']
    assert file_status['error_list'] == []


def test_diffmd5validate_check_existing_file_status():
    '''
    If md5 is different from stored md5, must re-validate file
    '''
    entity = synapseclient.Entity(name='first.txt', id='syn1234', md5='44444')
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf), mock_csv_query_result(emptydf), entities)
    assert file_status['to_validate']
    assert file_status['status_list'] == ['VALID']
    assert file_status['error_list'] == []


def test_diffnametovalidate_check_existing_file_status():
    '''
    If name is different from stored name, must re-validate file
    '''
    entity = synapseclient.Entity(name='second.txt', id='syn1234', md5='3333')
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf), mock_csv_query_result(emptydf), entities)
    assert file_status['to_validate']
    assert file_status['status_list'] == ['VALID']
    assert file_status['error_list'] == []


def test_twoinvalid_check_existing_file_status():
    '''
    Test to make sure both invalid statues get passed back
    '''
    validation_statusdf = pd.DataFrame({
        'id': ['syn1234', 'syn2345'],
        'status': ['INVALID', 'INVALID'],
        'md5': ['3333', '44444'],
        'name': ['first.txt', 'second.txt']})
    error_trackerdf = pd.DataFrame({
        'id': ['syn1234', 'syn2345'],
        'errors': ['Invalid file format', 'Invalid formatting issues']})
    first_entity = synapseclient.Entity(name='first.txt', id='syn1234', md5='3333')
    second_entity = synapseclient.Entity(name='second.txt', id='syn2345', md5='44444')
    entities = [first_entity, second_entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf), mock_csv_query_result(error_trackerdf), entities)
    assert not file_status['to_validate']
    assert file_status['status_list'] == [
        'INVALID', 'INVALID']
    assert file_status['error_list'] == [
        'Invalid file format', 'Invalid formatting issues']


def test_error_check_existing_file_status():
    '''
    With exception of clinical files, there should never be
    more than 2 files being passed in for validation.
    '''
    with pytest.raises(
            ValueError,
            match='There should never be more than 2 files being validated.'):
        entities = ['foo', 'doo', 'boo']
        input_to_database.check_existing_file_status(
            emptydf, emptydf, entities)


def test_valid_validatefile():
    '''
    Tests the behavior of a file that gets validated that becomes
    valid
    '''
    validation_statusdf = pd.DataFrame()
    error_trackerdf = pd.DataFrame()
    entity = synapseclient.Entity(name="data_clinical_supp_SAGE.txt",
                                  id='syn1234', md5='44444',
                                  path='/path/to/data_clinical_supp_SAGE.txt')
    entity['modifiedOn'] = '2019-03-24T12:00:00.Z'
    # This modifiedOn translates to: 1553428800000
    entity.modifiedBy = '333'
    entity.createdBy = '444'
    entities = [entity]
    threads = 0
    valid = True
    message = "Is valid"
    filetype = "clinical"
    # Only a list is returned as opposed a list of lists when there are
    # invalid errors
    status_error_list_results = ([{'entity': entity, 'status': 'VALIDATED'}],
                                 [])
    expected_results = ([{'entity': entity, 'status': 'VALIDATED',
                          'fileType': filetype, 'center': center}],
                        [], [])
    with patch.object(GenieValidationHelper, "determine_filetype",
                      return_value=filetype) as patch_determine_filetype,\
         patch.object(input_to_database, "check_existing_file_status",
                      return_value={'status_list': [],
                                    'error_list': [],
                                    'to_validate': True}) as patch_check, \
         patch.object(GenieValidationHelper, "validate_single_file",
                      return_value=(valid, message)) as patch_validate,\
         patch.object(input_to_database, "_get_status_and_error_list",
                      return_value=status_error_list_results) as patch_get_staterror_list,\
         patch.object(input_to_database,
                      "_send_validation_error_email") as patch_send_email:

        validate_results = input_to_database.validatefile(
            syn, None, entities, validation_statusdf,
            error_trackerdf, center, threads, oncotree_link)

        assert expected_results == validate_results
        patch_validate.assert_called_once_with(
            oncotree_link=oncotree_link, nosymbol_check=False)
        patch_check.assert_called_once_with(
            validation_statusdf, error_trackerdf, entities)
        patch_determine_filetype.assert_called_once()
        patch_get_staterror_list.assert_called_once_with(
            valid, message, entities)
        patch_send_email.assert_not_called()


def test_invalid_validatefile():
    '''
    Tests the behavior of a file that gets validated that becomes
    invalid
    '''
    validation_statusdf = pd.DataFrame()
    error_trackerdf = pd.DataFrame(columns=['id'], dtype=str)
    entity = synapseclient.Entity(name="data_clinical_supp_SAGE.txt",
                                  id='syn1234', md5='44444',
                                  path='/path/to/data_clinical_supp_SAGE.txt')
    entity['modifiedOn'] = '2019-03-24T12:00:00.Z'
    # This modifiedOn translates to: 1553428800000
    entity.modifiedBy = '333'
    entity.createdBy = '444'
    entities = [entity]
    threads = 0
    valid = False
    message = "Is invalid"
    filetype = "clinical"
    status_error_list_results = ([{'entity': entity, 'status': 'INVALID'}],
                                 [{'entity': entity, 'errors': message}])
    expected_results = ([{'entity': entity, 'status': 'INVALID',
                          'fileType': filetype, 'center': center}],
                        [{'entity': entity, 'errors': message,
                          'fileType': filetype, 'center': center}],
                          [(['data_clinical_supp_SAGE.txt'], 'Is invalid', ['333', '444'])])

    with patch.object(GenieValidationHelper, "determine_filetype",
                      return_value=filetype) as patch_determine_filetype,\
         patch.object(input_to_database, "check_existing_file_status",
                      return_value={'status_list': [],
                                    'error_list': [],
                                    'to_validate': True}) as patch_check, \
         patch.object(GenieValidationHelper, "validate_single_file",
                      return_value=(valid, message)) as patch_validate,\
         patch.object(input_to_database, "_get_status_and_error_list",
                      return_value=status_error_list_results) as patch_get_staterror_list:

        validate_results = input_to_database.validatefile(
            syn, None, entities, validation_statusdf,
            error_trackerdf, center, threads, oncotree_link
        )

        assert expected_results == validate_results
        patch_validate.assert_called_once_with(
            oncotree_link=oncotree_link, nosymbol_check=False)
        patch_check.assert_called_once_with(
            validation_statusdf, error_trackerdf, entities)
        patch_determine_filetype.assert_called_once()
        patch_get_staterror_list.assert_called_once_with(
            valid, message, entities)


def test_already_validated_validatefile():
    '''
    Test already validated files
    '''
    validation_statusdf = pd.DataFrame()
    error_trackerdf = pd.DataFrame()
    entity = synapseclient.Entity(name="data_clinical_supp_SAGE.txt",
                                  id='syn1234', md5='44444',
                                  path='/path/to/data_clinical_supp_SAGE.txt')
    entity['modifiedOn'] = '2019-03-24T12:00:00.Z'
    # This modifiedOn translates to: 1553428800000
    entity.modifiedBy = '333'
    entity.createdBy = '444'
    entities = [entity]
    threads = 0
    valid = False
    errors = "Invalid file"
    filetype = "markdown"
    status = "INVALID"
    check_file_status_dict = {
        'status_list': [status],
        'error_list': [errors],
        'to_validate': False}

    status_error_list_results = ([{'entity': entity, 'status': status}],
                                 [{'entity': entity, 'errors': errors}])
    expected_results = ([{'entity': entity, 'status': status,
                          'fileType': filetype, 'center': center}],
                        [{'entity': entity, 'errors': errors,
                          'fileType': filetype, 'center': center}],
                        [])
    with patch.object(GenieValidationHelper, "determine_filetype",
                      return_value=filetype) as patch_determine_filetype,\
         patch.object(input_to_database, "check_existing_file_status",
                      return_value=check_file_status_dict) as patch_check, \
         patch.object(GenieValidationHelper, "validate_single_file",
                      return_value=(valid, errors)) as patch_validate,\
         patch.object(input_to_database, "_get_status_and_error_list",
                      return_value=status_error_list_results) as patch_get_staterror_list,\
         patch.object(input_to_database,
                      "_send_validation_error_email") as patch_send_email:

        validate_results = input_to_database.validatefile(
            syn, None, entities, validation_statusdf,
            error_trackerdf, center, threads, oncotree_link,
        )

        assert expected_results == validate_results
        patch_validate.assert_not_called()
        patch_check.assert_called_once_with(
            validation_statusdf, error_trackerdf, entities)
        patch_determine_filetype.assert_called_once()
        patch_get_staterror_list.assert_not_called()
        patch_send_email.assert_not_called()


def test_valid__get_status_and_error_list():
    '''
    Tests the correct status and error lists received
    when file is valid.
    '''
    modified_on = 1561143558000
    modified_on_string = "2019-06-21T18:59:18.456Z"

    entity = synapseclient.Entity(id='syn1234', md5='44444',
                                  path='/path/to/foobar.txt',
                                  name='data_clinical_supp_SAGE.txt')
    entity.properties.modifiedOn = modified_on_string

    entities = [entity]

    valid = True
    message = 'valid'
    filetype = 'clinical'

    input_status_list, invalid_errors_list = \
        input_to_database._get_status_and_error_list(
           valid, message, entities)
    assert input_status_list == [{'entity': entity, 'status': 'VALIDATED'}]
    assert not invalid_errors_list


def test_invalid__get_status_and_error_list():
    '''
    Tests the correct status and error lists received
    when file is invalid.
    '''
    modified_on = 1561143558000
    modified_on_string = "2019-06-21T18:59:18.456Z"
    entity = synapseclient.Entity(id='syn1234', md5='44444',
                                  path='/path/to/foobar.txt',
                                  name='data_clinical_supp_SAGE.txt')
    entity.properties.modifiedOn = modified_on_string

    entities = [entity]
    filetype = "clinical"
    # This valid variable control the validation status
    valid = False
    errors = 'invalid file content'

    input_status_list, invalid_errors_list = \
        input_to_database._get_status_and_error_list(
            valid, errors, entities)
    assert input_status_list == [{'entity': entity, 'status': 'INVALID'}]
    assert invalid_errors_list == [{'entity': entity, 'errors': errors}]


def test__send_validation_error_email():
    message = "invalid error message here"
    filenames = ['data_clinical_supp_SAGE.txt']
    file_user = '333'
    message_objs = [dict(filenames=filenames, messages=message)]
    with patch.object(syn, "getUserProfile",
                      return_value={'userName':
                                    'trial'}) as patch_syn_getuserprofile,\
         patch.object(syn, "sendMessage") as patch_syn_sendmessage,\
         patch('genie.input_to_database.datetime') as mock_datetime:
        mock_datetime.datetime.today.return_value = datetime(2019, 11, 1, 00,
                                                             00, 00, 0)
        mock_datetime.side_effect = lambda *args, **kw: date(*args, **kw)

        input_to_database._send_validation_error_email(syn, file_user,
                                                       message_objs)
        error_message = ("Dear trial,\n\n"
                         "You have invalid files! "
                         "Here are the reasons why:\n\n"
                         "Filenames: data_clinical_supp_SAGE.txt, "
                         f"Errors:\n {message}\n\n")
        patch_syn_sendmessage.assert_called_once_with(
            userIds=[file_user], messageBody=error_message,
            messageSubject='GENIE Validation Error - 2019-11-01 00:00:00')
        patch_syn_getuserprofile.assert_called_once_with(file_user)


class emptytable_mock:
    '''
    Validation status tablequery dataframe mocking and
    error tracking tablequery dataframe mocking
    This is used because assert_called_once_with has a hard
    time with comparing pandas dataframes
    '''
    tableId = "syn555"

    def asDataFrame(self):
        return([])


class TestValidation:
    def setup(self):
        valid = [[sample_clinical_entity.id, sample_clinical_entity.path,
                  sample_clinical_entity.md5, 'VALIDATED',
                  sample_clinical_entity.name, 1553428800000, 'clinical',
                  center, sample_clinical_entity]]
        self.empty_validation = pd.DataFrame(
            columns=["id", 'path', 'md5', 'status', 'name',
                     'modifiedOn', 'fileType', 'center', 'entity']
        )
        self.validation_statusdf = pd.DataFrame(
            valid, columns=["id", 'path', 'md5', 'status', 'name',
                            'modifiedOn', 'fileType', 'center', 'entity']
        )
        error = [[sample_clinical_entity.id, "errors",
                  sample_clinical_entity.name,
                  'clinical', center, sample_clinical_entity]]
        self.errors_df = pd.DataFrame(
            error,
            columns=['id', 'errors', 'name', 'fileType', 'center', 'entity']
        )
        self.empty_errors = pd.DataFrame(
            columns=['id', 'errors', 'name', 'fileType', 'center', 'entity']
        )

        self.with_dupsdf = pd.DataFrame({
            'id': [sample_clinical_entity.id,
                   'syn2345', 'syn5555', 'syn1224', 'syn34444'],
            'name': ['first.cbs', 'second.seg', 'data_clinical_supp_1',
                     'data_clinical_supp_2', 'data_clinical_supp_3'],
            'center': ['SAGE']*5,
            'fileType': ['type']*5,
            'entity': ['entity']*5
        })
        self.duplicateddf = self.with_dupsdf.copy()
        self.duplicateddf['errors'] = input_to_database.DUPLICATED_FILE_ERROR

        self.no_dupsdf = pd.DataFrame({
            'id': [sample_clinical_entity.id,
                   'syn2345', 'syn5555', 'syn1224', 'syn34444'],
            'name': ['cbs.txt', 'second.seg', 'no_clinical.txt',
                     'data_clinical_supp_2', 'data_clinical_supp_3'],
            'center': ['SAGE']*5,
            'fileType': ['type']*5,
            'entity': ['entity']*5
        })
        self.empty_dup = pd.DataFrame(columns=['id', 'name', 'center',
                                               'fileType', 'entity',
                                               'errors'])

    def test_build_validation_status_table(self):
        input_valid_statuses = [{'entity': sample_clinical_entity,
                                 'status': 'VALIDATED',
                                 'fileType': 'clinical',
                                 'center': center}]
        validationdf = input_to_database.build_validation_status_table(
            input_valid_statuses
        )
        assert validationdf.equals(self.validation_statusdf)

    def test_build_validation_status_table__empty(self):
        input_valid_statuses = []
        validationdf = input_to_database.build_validation_status_table(
            input_valid_statuses
        )
        assert validationdf.equals(self.empty_validation)

    def test_build_error_tracking_table(self):
        invalid_errors = [{'entity': sample_clinical_entity,
                           'errors': 'errors',
                           'fileType': 'clinical',
                           'center': center}]
        errordf = input_to_database.build_error_tracking_table(
            invalid_errors
        )
        assert errordf.equals(self.errors_df)

    def test_build_error_tracking_table__empty(self):
        invalid_errors = []
        errordf = input_to_database.build_error_tracking_table(
            invalid_errors
        )
        assert errordf.equals(self.empty_errors)

    def test_update_status_and_error_tables(self):
        """Test updating validation status and error table"""
        validation_status_table = emptytable_mock()
        error_tracker_table = emptytable_mock()

        with patch.object(process_functions, "updateDatabase") as mock_update:
            input_to_database.update_status_and_error_tables(
                syn,
                self.validation_statusdf,
                self.errors_df,
                validation_status_table,
                error_tracker_table
            )
            assert mock_update.call_count == 2

    def test_dups_get_duplicated_files(self):
        """Test get all duplicates: cbs/seg/clinical"""
        dupsdf = input_to_database.get_duplicated_files(self.with_dupsdf)
        assert dupsdf.equals(self.duplicateddf)

    def test_nodups_get_duplicated_files(self):
        """Test no duplicated"""
        dupsdf = input_to_database.get_duplicated_files(self.no_dupsdf)
        assert dupsdf.equals(self.empty_dup)

    def test__update_tables_content(self):
        """Tests duplicates are added to the tables and errors/statues are
        updated
        """
        errorsdf = self.errors_df.copy()
        errorsdf['errors'] = input_to_database.DUPLICATED_FILE_ERROR

        validationdf = self.validation_statusdf
        validationdf['status'] = 'INVALID'

        with patch.object(input_to_database,
                          "get_duplicated_files",
                          return_value=self.duplicateddf):
            updated_tables = input_to_database._update_tables_content(
                self.validation_statusdf,
                self.errors_df
            )
        assert updated_tables['duplicated_filesdf'].equals(self.duplicateddf)
        assert updated_tables['error_trackingdf'].equals(errorsdf)
        assert updated_tables['validation_statusdf'].equals(validationdf)

    def test__update_tables_content__remove_old_duplicates(self):
        """Tests that old duplicates are removed"""
        errorsdf = self.errors_df.copy()
        errorsdf['errors'] = input_to_database.DUPLICATED_FILE_ERROR
        validationdf = self.validation_statusdf
        validationdf['status'] = 'INVALID'

        with patch.object(input_to_database,
                          "get_duplicated_files",
                          return_value=self.empty_dup):
            updated_tables = input_to_database._update_tables_content(
                validationdf,
                errorsdf
            )
        assert updated_tables['duplicated_filesdf'].empty
        assert updated_tables['error_trackingdf'].empty
        assert updated_tables['validation_statusdf'].empty

    def test__update_tables_content__remove_old_errors(self):
        """Tests that old errors are removed"""
        errorsdf = self.errors_df.copy()
        errorsdf['errors'] = input_to_database.DUPLICATED_FILE_ERROR
        validationdf = self.validation_statusdf
        validationdf['status'] = 'VALIDATED'

        with patch.object(input_to_database,
                          "get_duplicated_files",
                          return_value=self.empty_dup):
            updated_tables = input_to_database._update_tables_content(
                validationdf,
                errorsdf
            )
        assert updated_tables['duplicated_filesdf'].empty
        assert updated_tables['error_trackingdf'].empty
        assert updated_tables['validation_statusdf'].empty

    def test_validation(self):
        """Test validation steps"""
        modified_on = 1561143558000
        process = "main"
        databaseToSynIdMapping = {'Database': ["clinical", 'validationStatus', 'errorTracker'],
                                'Id': ['syn222', 'syn333', 'syn444']}
        databaseToSynIdMappingDf = pd.DataFrame(databaseToSynIdMapping)
        entity = synapseclient.Entity(id='syn1234', md5='44444',
                                      path='/path/to/foobar.txt',
                                      name='data_clinical_supp_SAGE.txt')
        entities = [entity]
        filetype = "clinical"
        input_status_list = [[entity.id, entity.path, entity.md5,
                              'VALIDATED', entity.name, modified_on,
                              filetype, center]]
        invalid_errors_list = []
        messages = []
        new_tables = {'validation_statusdf': self.validation_statusdf,
                      'error_trackingdf': self.errors_df,
                      'duplicated_filesdf': self.empty_dup}
        validationstatus_mock = emptytable_mock()
        errortracking_mock = emptytable_mock()
        valiate_cls = Mock()
        with patch.object(syn, "tableQuery",
                          side_effect=[validationstatus_mock,
                                       errortracking_mock]) as patch_query,\
             patch.object(input_to_database, "validatefile",
                          return_value=(input_status_list,
                                        invalid_errors_list,
                                        messages)) as patch_validatefile,\
             patch.object(input_to_database, "build_validation_status_table",
                          return_value=self.validation_statusdf),\
             patch.object(input_to_database, "build_error_tracking_table",
                          return_value=self.errors_df),\
             patch.object(input_to_database, "_update_tables_content",
                          return_value=new_tables),\
             patch.object(input_to_database, "update_status_and_error_tables"):

            valid_filedf = input_to_database.validation(
                syn, "syn123", center, process,
                entities, databaseToSynIdMappingDf,
                oncotree_link,
                format_registry={"test": valiate_cls}
            )
            assert patch_query.call_count == 2
            patch_validatefile.assert_called_once_with(
                syn, "syn123", entity,
                validationstatus_mock,
                errortracking_mock,
                center='SAGE', threads=1,
                oncotree_link=oncotree_link,
                format_registry={"test": valiate_cls}
            )

            assert valid_filedf.equals(
                self.validation_statusdf[['id', 'path', 'fileType', 'name']]
            )


@pytest.mark.parametrize(
    'process, genieclass, filetype', [
        ('main', Mock(), 'clinical'),
        ('main', Mock(), 'maf'),
    ]
)
def test_main_processfile(process, genieclass, filetype):
    validfiles = {'id': ['syn1'],
                  'path': ['/path/to/data_clinical_supp.txt'],
                  'fileType': [filetype],
                  'name': ['data_clinical_supp_SAGE.txt']}
    validfilesdf = pd.DataFrame(validfiles)
    center = "SAGE"
    path_to_genie = "./"
    oncotree_link = "www.google.com"
    center_mapping = {'stagingSynId': ["syn123"],
                      'center': [center]}
    center_mapping_df = pd.DataFrame(center_mapping)
    databaseToSynIdMapping = {'Database': [filetype],
                              'Id': ['syn222']}
    databaseToSynIdMappingDf = pd.DataFrame(databaseToSynIdMapping)
    format_registry = {filetype: genieclass}

    input_to_database.processfiles(
        syn, validfilesdf, center, path_to_genie,
        center_mapping_df, oncotree_link, databaseToSynIdMappingDf,
        processing=process,
        format_registry=format_registry
    )
    genieclass.assert_called_once()



def test_mainnone_processfile():
    """If file type is None, the processing function is not called"""
    validfiles = {'id': ['syn1'],
                  'path': ['/path/to/data_clinical_supp_SAGE.txt'],
                  'fileType': [None],
                  'name': ['data_clinical_supp_SAGE.txt']}
    validfilesdf = pd.DataFrame(validfiles)
    center = "SAGE"
    path_to_genie = "./"
    oncotree_link = "www.google.com"
    center_mapping = {'stagingSynId': ["syn123"],
                      'center': [center]}
    center_mapping_df = pd.DataFrame(center_mapping)
    databaseToSynIdMapping = {'Database': ["clinical"],
                              'Id': ['syn222']}
    databaseToSynIdMappingDf = pd.DataFrame(databaseToSynIdMapping)
    process_cls = Mock()

    with patch.object(clinical, "process") as patch_clin:
        input_to_database.processfiles(
            syn, validfilesdf, center, path_to_genie,
            center_mapping_df, oncotree_link, databaseToSynIdMappingDf,
            processing="main",
            format_registry={"main": process_cls})
        patch_clin.assert_not_called()


def test_mutation_processfile():
    '''
    Make sure mutation is called correctly
    '''
    validfiles = {'id': ['syn1'],
                  'path': ['/path/to/data_clinical_supp_SAGE.txt'],
                  'fileType': [None]}
    validfilesdf = pd.DataFrame(validfiles)
    center = "SAGE"
    path_to_genie = "./"
    oncotree_link = "www.google.com"
    center_mapping = {'stagingSynId': ["syn123"],
                      'center': [center]}
    center_mapping_df = pd.DataFrame(center_mapping)
    databaseToSynIdMapping = {'Database': ['vcf'],
                              'Id': ['syn222']}
    databaseToSynIdMappingDf = pd.DataFrame(databaseToSynIdMapping)
    process_cls = Mock()

    with patch.object(process_mutation,
                      "process_mutation_workflow") as patch_process:
        input_to_database.processfiles(
            syn, validfilesdf, center, path_to_genie,
            center_mapping_df, oncotree_link, databaseToSynIdMappingDf,
            processing='mutation',
            genome_nexus_pkg="/path/to/nexus",
            format_registry={"vcf": process_cls})
        # TODO: fix hardcoding
        patch_process.assert_called_once_with(
            syn=syn,
            center=center,
            validfiles=validfilesdf,
            genie_annotation_pkg="/path/to/nexus",
            database_mappingdf=databaseToSynIdMappingDf,
            workdir=path_to_genie
        )
